/*###########################################################################################################################
# Copyright (c) 1997-2012 Ufasoft   http://ufasoft.com   mailto:support@ufasoft.com                                         #
#                                                                                                                           #
# This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License #
# as published by the Free Software Foundation; either version 3, or (at your option) any later version.                    #                                                          #
#                                                                                                                           #
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied        #
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.     #
#                                                                                                                           #
# You should have received a copy of the GNU General Public License along with this program;                                #
# If not, see <http://www.gnu.org/licenses/>                                                                                #
###########################################################################################################################*/

#pragma once

#if UCFG_WIN32
#	include <winternl.h>
#endif

#if UCFG_USE_PTHREADS
#	include <semaphore.h>
#endif

#if !UCFG_STDSTL || !UCFG_CPP11_HAVE_MUTEX
#	include <el/stl/mutex>

namespace std {
	using ExtSTL::mutex;
	using ExtSTL::lock_guard;
}

#endif


namespace Ext {

#if UCFG_USE_PTHREADS
void PthreadCheck(int code);
#endif

#ifdef _WIN32
NTSTATUS AFXAPI NtCheck(NTSTATUS status, NTSTATUS allowableError = 0);		// 0 == STATUS_SUCCESS
#endif

class AFX_CLASS COperatingSystem {
	typedef COperatingSystem class_type;
public:

#if UCFG_WDM
	typedef RTL_OSVERSIONINFOEXW COsVerInfo;
#elif UCFG_WIN32
	typedef OSVERSIONINFOEX COsVerInfo;
#endif

	static String AFXAPI get_ComputerName();
	DEFPROP_GET(String, ComputerName);

	static String AFXAPI get_UserName();
	DEFPROP_GET(String, UserName);

#if UCFG_WIN32 || UCFG_USE_POSIX
	static String AFXAPI get_ExeFilePath();
	DEFPROP_GET(String, ExeFilePath);
#endif

#ifdef _WIN32
	EXT_API static COsVerInfo AFXAPI get_Version();
	DEFPROP_GET(COsVerInfo, Version);
#endif

#if UCFG_WIN32
	static DWORD AFXAPI GetSysColor(int nIndex);
	static void AFXAPI MessageBeep(UINT uType = 0xFFFFFFFF);
	EXT_API static CStringVector AFXAPI QueryDosDevice(RCString dev);


	static SYSTEM_INFO get_Info() {
		SYSTEM_INFO si;
		::GetSystemInfo(&si);
		return si;
	}
	DEFPROP_GET(SYSTEM_INFO, Info);

	bool Is64BitNativeSystem();

	static String AFXAPI get_WindowsDirectory();
	DEFPROP_GET(String, WindowsDirectory);

#	if UCFG_WIN32_FULL
	EXT_API static std::vector<String> AFXAPI get_LogicalDriveStrings();
	DEFPROP_GET(std::vector<String>, LogicalDriveStrings);
#	endif

#endif

	static Int64 AFXAPI get_PerformanceCounter();
	DEFPROP_GET(Int64, PerformanceCounter);

	static Int64 AFXAPI get_PerformanceFrequency();
	DEFPROP_GET(Int64, PerformanceFrequency);
};

extern EXT_DATA COperatingSystem System;


class AFX_CLASS CSyncObject : public SafeHandle {
	typedef CSyncObject class_type;
public:
	CBool m_bAlreadyExists;

	CSyncObject() {}
	CSyncObject(RCString pstrName);
	void AttachCreated(HANDLE h);
	virtual void Unlock() =0;

	virtual bool Lock(DWORD dwTimeout = INFINITE);
	void Unlock(Int32 lCount, Int32 *lprevCount = 0);
#ifdef WDM_DRIVER
	bool LockEx(DWORD dwTimeout, KPROCESSOR_MODE mode, bool bAlertable);
#endif
private:
	CSyncObject(const class_type&);
};

class AFX_CLASS CCriticalSection : public CSyncObject {
public:
	CCriticalSection();
	//  CCriticalSection(DWORD dwSpinCount);
	~CCriticalSection();

	void lock() {
#if UCFG_USE_PTHREADS
		PthreadCheck(::pthread_mutex_lock(&m_mutex));
#elif defined(WDM_DRIVER)
		KeAcquireSpinLock(&m_spinLock, &m_oldIrql);
#else
		::EnterCriticalSection(&m_sect);				// don't handle EXCEPTION_POSSIBLE_DEADLOCK and "OutOfMemory on Win2K"
#endif
	}

	void unlock() {
#if UCFG_USE_PTHREADS
		PthreadCheck(::pthread_mutex_unlock(&m_mutex));
#elif defined(WDM_DRIVER)
		KeReleaseSpinLock(&m_spinLock, m_oldIrql);
#else
		::LeaveCriticalSection(&m_sect);
#endif
	}

	bool Lock(DWORD dwTimeout = INFINITE);
	bool TryLock();
	void Unlock();
	
	//!!!  DWORD SetSpinCount(DWORD dwSpinCount);
private:

#if UCFG_USE_PTHREADS
	pthread_mutex_t m_mutex;
#elif defined WDM_DRIVER
    KSPIN_LOCK m_spinLock;
	KIRQL m_oldIrql;
#else
	CRITICAL_SECTION m_sect;
#endif
	void Init();

friend class CEvent;
};

class CTryLock {
public:
	CTryLock(CCriticalSection& cs) {
		m_pCs = cs.TryLock() ? &cs : 0; 
	}

	~CTryLock() {
		if (m_pCs)
			m_pCs->Unlock();
	}

	struct _Boolean {
		int i;
	};

	operator int _Boolean::*() const {
		return m_pCs ? &_Boolean::i : 0;
	}
private:
	CCriticalSection *m_pCs;
};

class CEvent : public CSyncObject {
	typedef CSyncObject base;
public:
	CEvent(std::nullptr_t p) {
	}
#if UCFG_USE_PTHREADS
	CEvent(bool bInitiallyOwn = false, bool bManualReset = false);
	void Set();
#else
	CEvent(bool bInitiallyOwn = false, bool bManualReset = false, RCString name = nullptr
#ifdef WIN32
		, LPSECURITY_ATTRIBUTES lpsaAttribute = 0
#endif
		);
#endif
	~CEvent();

#if UCFG_WDM
	KEVENT *Obj() { return (KEVENT*)m_pObject; }
#endif

#if !UCFG_USE_PTHREADS
	void Set() {
#	ifdef WDM_DRIVER
		::KeSetEvent(Obj(), IO_NO_INCREMENT, FALSE);
#	else		
		Win32Check(::SetEvent(HandleAccess(*this)));
#	endif
	}
#endif

#if !UCFG_USE_PTHREADS && !defined(WDM_DRIVER)
	void Pulse() { Win32Check(::PulseEvent(HandleAccess(*this))); }
#endif

	void Reset();
	bool Lock(DWORD dwTimeout = INFINITE);
	void Unlock();
private:
#if UCFG_USE_PTHREADS
	pthread_cond_t m_cond;
	CCriticalSection m_mutex;
	CBool m_bManual, m_bSignaled;
#elif UCFG_WDM
	KEVENT m_ev;
#endif
};

typedef CEvent EventWaitHandle;

class AutoResetEvent : public EventWaitHandle {
	typedef EventWaitHandle base;
public:
	AutoResetEvent(bool bInit = false)
		:	base(bInit, false)
	{}
};

class ManualResetEvent : public EventWaitHandle {
	typedef EventWaitHandle base;
public:
	ManualResetEvent(bool bInit = false)
		:	base(bInit, true)
	{}
};


class CMutex : public CSyncObject {
public:
	CMutex(bool bInitiallyOwn = false, RCString name = nullptr
#ifdef WIN32
		, LPSECURITY_ATTRIBUTES lpsaAttribute = 0
#endif
	);
#if UCFG_WIN32_FULL
	CMutex(RCString name, DWORD dwAccess = MUTEX_ALL_ACCESS, bool bInherit = false);
#endif
	~CMutex();

#if UCFG_WDM
	KMUTEX *Obj() { return (KMUTEX*)m_pObject; }
#endif

	void Unlock();
private:
#if UCFG_USE_PTHREADS
#elif UCFG_WDM
	KMUTEX m_mutex;
#endif
};

class AFX_CLASS CSemaphore : public CSyncObject {
public:
	CSemaphore(LONG lInitialCount = 1, LONG lMaxCount = 1, RCString name = nullptr
#ifdef WIN32		
		, LPSECURITY_ATTRIBUTES lpsaAttributes = 0
#endif
		);
#if UCFG_WIN32_FULL
	CSemaphore(RCString name, DWORD dwAccess = SEMAPHORE_ALL_ACCESS, bool bInherit = false);
#endif
	~CSemaphore();

#if UCFG_USE_PTHREADS
	sem_t m_sem, *m_psem;
#endif

#if UCFG_WDM
	KSEMAPHORE m_sem;
	KSEMAPHORE *Obj() { return (KSEMAPHORE*)m_pObject; }
#endif

#if UCFG_USE_PTHREADS
	bool Lock(DWORD dwTimeout = INFINITE) {
		while (true) {
			int rc = ::sem_wait(m_psem);
			if (!rc)
				return true;
			if (errno != EINTR)
				CCheck(-1);				
		}
	}
#endif

	void Unlock();
	LONG Unlock(LONG lCount);
};

/*!!!R
class CLock {
public:
	CLock(CSyncObject& sync)
		:	m_sync(sync)
	{
		m_sync.Lock();
	}

	~CLock() {
		m_sync.Unlock();
	}
private:
	CSyncObject& m_sync;
};*/

template <class T> void Lock(T& sync) {
	sync.Lock();
}

template <class T> void Unlock(T& sync) {
	sync.Unlock();
}

inline void Lock(CCriticalSection& cs) {
	cs.lock();
}

inline void Unlock(CCriticalSection& cs) {
	cs.unlock();
}

#if UCFG_STDSTL && UCFG_CPP11_HAVE_MUTEX

inline void Lock(std::mutex& mtx) {
	mtx.lock();
}

inline void Unlock(std::mutex& mtx) {
	mtx.unlock();
}

#endif // UCFG_STDSTL && UCFG_CPP11_HAVE_MUTEX


class CScopedLockBase {
public:
	operator bool() const {
		return false;
	}
};

template <class T>
class CScopedLock : public CScopedLockBase {
public:
	mutable T *m_sync;
	
	CScopedLock(const CScopedLock& sl)
		:	m_sync(SwapRet(sl.m_sync, nullptr))
	{}

	CScopedLock(T& sync)
		:	m_sync(&sync)
	{
		Lock(*m_sync);
	}
	
	~CScopedLock() {
		if (m_sync)
			Ext::Unlock(*m_sync);
	}


/*!!!	void Unlock() const {
		Ext::Unlock(*m_sync);
		m_sync = 0;
	}*/
};


template <class T> CScopedLock<T> ScopedLock(T& sync) { return CScopedLock<T>(sync); }

#define EXT_LOCK(m) if (const Ext::CScopedLockBase& lck = Ext::ScopedLock(m)) Throw(E_EXT_CodeNotReachable); else
//#define EXT_LOCK(m) if (const Ext::CScopedLockBase& lck = Ext::ScopedLock(m)) ; else

#define EXT_LOCKED(cs, expr) (std::lock_guard<decltype(cs)>(cs), (expr))			//!!!

#if UCFG_SIMPLE_MACROS
#	define lock EXT_LOCK
#endif

} // Ext::





