/*###########################################################################################################################
# Copyright (c) 1997-2012 Ufasoft   http://ufasoft.com   mailto:support@ufasoft.com                                         #
#                                                                                                                           #
# This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License #
# as published by the Free Software Foundation; either version 3, or (at your option) any later version.                    #                                                          #
#                                                                                                                           #
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied        #
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.     #
#                                                                                                                           #
# You should have received a copy of the GNU General Public License along with this program;                                #
# If not, see <http://www.gnu.org/licenses/>                                                                                #
###########################################################################################################################*/

#include <el/ext.h>

#if UCFG_EXTENDED
#	if !UCFG_WCE
#		include <eh.h>
#	endif
#	include <el/gui/controls.h>
#endif
//!!!#include <exception>//!!!

namespace Ext {

using namespace std;

#if UCFG_EXTENDED
#	include "win32/mfcimpl.h"
#endif

/*!!!R
CCriticalSection g_criticalSection,
                 g_criticalSection2,
                 g_criticalSection3;*/

ThreadBase::ThreadBase(CThreadRef *ownRef)
	:	//!!! CWinThread(ThreaderFunction, this),
		StackSize(0)
	,	m_bAutoDelete(true)
	,	m_owner(ownRef)
	,	m_bStop(false)
	,	m_bSleeping(false)
	,	m_bAPC(false)
#if UCFG_USE_PTHREADS
		,	m_tid()
#elif UCFG_OLE
		,	m_usingCOM(E_NoInit)
		,	m_dwCoInit(COINIT_APARTMENTTHREADED)
#endif
{
	ASSERT(1000 < std::abs(Int32((byte*)&ownRef - (byte*)this)));		// Thread cannot be in stack
}

ThreadBase::~ThreadBase() {
	if (m_threadRef) {
		m_threadRef->StopChilds();
		delete m_threadRef;
	}


#if UCFG_EXTENDED
	// cleanup module state
	AFX_MODULE_THREAD_STATE* pState = AfxGetModuleThreadState();
	if (pState->m_pCurrentThread == this)
		pState->m_pCurrentThread = NULL;
#endif
}

//!!!DCTls Thread::t_pCurThreader;//!!!

//!!!#if UCFG_EXTENDED

ThreadBase::propclass_CurrentThread ThreadBase::CurrentThread; //!!!

#if !UCFG_EXTENDED
static CTls s_currentThread;
#endif

ThreadBase* ThreadBase::get_CurrentThread() {
#if UCFG_EXTENDED || UCFG_WIN32
	AFX_MODULE_THREAD_STATE& mts = *AfxGetModuleState()->m_thread;
	if (!mts.m_pCurrentThread) { //!!!
		mts.m_pCurrentThread = new Thread;
		mts.m_pCurrentThread->AttachSelf();
	}
	return mts.m_pCurrentThread;
#else
	if (!s_currentThread.get_Value())
		s_currentThread.Value = new Thread; // Leak
	return (ThreadBase*)s_currentThread.get_Value();
#endif
	//!!!D	return *(Thread*)t_pCurThreader.Value;
}

//!!!#endif

void ThreadBase::Sleep(DWORD dwMilliseconds) {
#ifdef WIN32
	CurrentThread->SleepImp(dwMilliseconds); //!!!
#else
	::usleep(dwMilliseconds*1000);
#endif
}

#if UCFG_USE_PTHREADS

void PthreadCheck(int code) {
	if (code)
		Throw(HRESULT_FROM_C(code));
}

void ThreadBase::ReleaseHandle(HANDLE h) const {
	PthreadCheck(::pthread_detach(SwapRet(m_tid, pthread_t())));
}

#endif

void ThreadBase::AttachSelf() {
#if UCFG_WIN32
	HANDLE h;
	//!!!#if UCFG_WCE
	h = GetCurrentThread();
	//!!!#else
	//!!!		Win32Check(DuplicateHandle(GetCurrentProcess(), GetCurrentThread(),
	//!!!															GetCurrentProcess(), &h, 0, FALSE, 0));
	//!!!#endif
	Attach(h, ::GetCurrentThreadId());
	m_bOwn = false;
#else
	m_tid = ::pthread_self();
#endif
}

CThreadRef& ThreadBase::GetThreadRef() {
	if (!m_threadRef)
		m_threadRef = new CThreadRef;
	return *m_threadRef;
}

void ThreadBase::SleepImp(DWORD dwMilliseconds) {
#if UCFG_WIN32_FULL
	if (m_bAPC)
		m_bAPC = false;
	else if (!m_bStop)
		SleepEx(dwMilliseconds, TRUE);
#else
	usleep(dwMilliseconds*1000);
#endif
	if (m_bStop)
		Throw(E_EXT_ThreadStopped);
}


/*!!!
LPTOP_LEVEL_EXCEPTION_FILTER g_pfnUnhadledExceptionFilter;

void __stdcall MyExceptHandler(EXCEPTION_POINTERS *ep)
{
TRC(1, "MyExcept_handler");//!!!T
if (g_pfnUnhadledExceptionFilter)
g_pfnUnhadledExceptionFilter(ep);
else
{
AfxLogErrorCode(0xFFFFFFFF); //!!1
::MessageBox(0, "Fatal Error", "Critical Error!\nReport about it to " MANUFACTURER_EMAIL, MB_ICONEXCLAMATION); //!!!
}
ExitProcess(3);
}

void __cdecl Myterminate_handler()//!!!
{
__try
{
RaiseException(0x02345678, 0, 0, 0);
}
__except(MyExceptHandler(GetExceptionInformation()))
{}
AfxLogErrorCode(0xFFFFFFFF); //!!1
::MessageBox(0, "Fatal Error", "Critical Error!\nReport about it to " MANUFACTURER_EMAIL, MB_ICONEXCLAMATION); //!!!
ExitProcess(3);
}
*/

#if UCFG_OLE

void ThreadBase::InitCOM() {
	m_usingCOM.Initialize(m_dwCoInit);
}

#endif

CThreadRef::~CThreadRef() {
	StopChilds();
}

/*!!!
void CThreadRef::SafeStopChilds()
{
if (m_ar.Count != 0)
{ 
m_evFinish.Set();
bool bTimeOut = false;
for (int i=m_ar.Count-1; i>=0; i--) //!!! may be very long time
{
Thread *t = m_ar[i];
if (!bTimeOut && WaitForSingleObject(*t, INFINITE) != WAIT_OBJECT_0)
Throw(E_FAIL);//!!!
{
CLock lock(m_cs);
m_ar.Remove(t);
}
delete t;
}
}
}*/

void CThreadRef::StopChilds() {
	SignalStop();
	WaitStop();
	/*!!!
	if (m_ar.Count != 0)
	{
	m_evFinish.Set();
	bool bTimeOut = false;
	for (int i=m_ar.Count-1; i>=0; i--) //!!! may be very long time
	{
	Thread *t = m_ar[i];
	//!!!      if (!bTimeOut && WaitForSingleObject(*t, INFINITE) != WAIT_OBJECT_0)
	if (!bTimeOut)
	{
	HANDLE h = *t;
	while (true)
	{
	DWORD r = MsgWaitForMultipleObjects(1, &h, FALSE, THREAD_TIMEOUT, QS_SENDMESSAGE);
	if (r == WAIT_OBJECT_0+1)
	{
	MSG msg;
	while (PeekMessage(&msg, 0, 0, 0, PM_REMOVE))
	DispatchMessage(&msg);
	}
	else
	{
	bTimeOut = r != WAIT_OBJECT_0;
	break;
	}
	}
	}
	//!!!      if (!bTimeOut && WaitForSingleObject(*t, THREAD_TIMEOUT) != WAIT_OBJECT_0)
	//        bTimeOut = true;
	if (t->ExitCode == STILL_ACTIVE)
	t->Terminate();
	{
	CLock lock(m_cs);
	m_ar.Remove(t);
	}
	delete t;
	}
	}*/
}

void CThreadRef::SignalStop() {
	EXT_LOCK (m_cs) {
		for (int i=0; i<m_ar.size(); i++) {
			ThreadBase *t = m_ar[i];
			t->m_bAutoDelete = false;
			if (*t)
				t->Stop();
		}
	}
}

void CThreadRef::WaitStop() {
	CThreadColl ar;
	EXT_LOCK (m_cs) {
		ar.swap(m_ar);
	}
	for (CThreadColl::iterator i(ar.begin()), e(ar.end()); i!=e; ++i) {
		if ((*i)->Valid())
			(*i)->Join();

		//!!!    int ec = t->ExitCode;
	}
}

bool CThreadRef::StopChild(ThreadBase *t, int msTimeout) {
	EXT_LOCK (m_cs) {
		t->m_bAutoDelete = false;
		if (*t)
			t->Stop();
	}
	if (*t) {
		if (!t->Join(msTimeout))
			return false;
	}

	EXT_LOCK (m_cs) {
		m_ar.erase(std::remove(m_ar.begin(), m_ar.end(), t), m_ar.end());
	}
	return true;
}

/*!!!
void CThreadRef::SafeStopChild(Thread *t)
{
{
CLock lock(m_cs);
t->m_bAutoDelete = false;
}
t->Stop();
HANDLE h = *t;
while (true)
{
DWORD r = MsgWaitForMultipleObjects(1, &h, FALSE, INFINITE, QS_SENDMESSAGE);
if (r == WAIT_OBJECT_0+1)
{
MSG msg;
while (PeekMessage(&msg, 0, 0, 0, PM_REMOVE))
DispatchMessage(&msg);
}
else
{
if (r != WAIT_OBJECT_0)
Throw(E_FAIL);
CLock lock(m_cs);
m_ar.Remove(t);
break;
}
}
delete t;
}*/

void ThreadBase::Delete() {
	if (m_owner) {
		m_owner->ExternalRelease();
		EXT_LOCK (m_owner->m_cs) {
			if (m_bAutoDelete)
				m_owner->m_ar.erase(std::remove(m_owner->m_ar.begin(), m_owner->m_ar.end(), this), m_owner->m_ar.end());
		}
	}
//!!!R	if (m_bAutoDelete)
//!!!R		delete this;
}

void ThreadBase::Execute() {
}

void ThreadBase::Run() {
	try {
		DBG_LOCAL_IGNORE(E_EXT_ThreadStopped);

		Execute();
	} catch (RCExc e) {
		m_exitCode = e.HResult;
	}
	TRC(2, "ExitThread: " << ID << "\t" << Name);
}

DWORD ThreadBase::get_ExitCode() {
#if UCFG_USE_PTHREADS
	return m_exitCode;
#else
	DWORD r;
	Win32Check(::GetExitCodeThread(HandleAccess(_self), &r));
	return r;
#endif
}

void ThreadBase::put_Name(RCString name) {
	m_name = name;
#if UCFG_USE_PTHREADS && defined(HAVE_PTHREAD_SETNAME_NP)
	CCheck(pthread_setname_np(m_tid, name.Left(15)));
#elif UCFG_WIN32
	struct THREADNAME_INFO {
		DWORD dwType; // must be 0x1000
		LPCSTR szName; // pointer to name (in user addr space)
		DWORD dwThreadID; // thread ID (-1=caller thread)
		DWORD dwFlags; // reserved for future use, must be zero
	};

	THREADNAME_INFO info;
	info.dwType = 0x1000;
	info.szName = m_name;
	info.dwThreadID = ID;
	info.dwFlags = 0;
	__try {
		RaiseException( 0x406D1388, 0, sizeof(info)/sizeof(DWORD), (ULONG_PTR*)&info );
	} __except(EXCEPTION_CONTINUE_EXECUTION) {
	}
#endif
}

int ThreadBase::get_Priority() {
#if UCFG_USE_PTHREADS
	return 0;
#else
	int p = ::GetThreadPriority(HandleAccess(_self));
	Win32Check(p != THREAD_PRIORITY_ERROR_RETURN);
	return p;
#endif
}

void ThreadBase::put_Priority(int nPriority) {
#if !UCFG_USE_PTHREADS
	Win32Check(::SetThreadPriority(HandleAccess(_self), nPriority));
#endif
}

void ThreadBase::OnEnd() {
#if UCFG_OLE
	m_usingCOM.Uninitialize();
#endif
}


#if UCFG_USE_PTHREADS
TimeSpan ThreadBase::get_TotalProcessorTime() const {
	timespec ts;
	CCheck(::clock_gettime(CLOCK_THREAD_CPUTIME_ID, &ts));
	return ts;
}

#else // UCFG_USE_PTHREADS

CONTEXT ThreadBase::get_Context(DWORD contextFlags) {
	CONTEXT context;
	context.ContextFlags = contextFlags;
	Win32Check(::GetThreadContext(HandleAccess(_self), &context));
	return context;
}

void ThreadBase::put_Context(const CONTEXT& context) {
	Win32Check(::SetThreadContext(HandleAccess(_self), &context));
}

CTimesInfo ThreadBase::get_Times() const {
	CTimesInfo r;
	Win32Check(::GetThreadTimes(HandleAccess(_self), &r.m_tmCreation, &r.m_tmExit, &r.m_tmKernel, &r.m_tmUser));
	return r;
}

DWORD ThreadBase::Resume() {
	DWORD prevCount = ::ResumeThread(HandleAccess(_self));
	Win32Check(prevCount != 0xFFFFFFFF);
	return prevCount;
}

DWORD ThreadBase::Suspend() {
	DWORD prevCount = ::SuspendThread(HandleAccess(_self));
	Win32Check(prevCount != 0xFFFFFFFF);
	return prevCount;
}

void ThreadBase::PostMessage(UINT message, WPARAM wParam, LPARAM lParam) {
	Win32Check(::PostThreadMessage(ID, message, wParam, lParam));
}

#	if !UCFG_WCE
void ThreadBase::OnAPC() {
	m_bAPC = true;
//!!!	if (m_bStop)
//!!!		Socket::ReleaseFromAPC();
}

VOID CALLBACK ThreadBase::APCProc(ULONG_PTR dwParam) {
	((ThreadBase*)dwParam)->OnAPC();
}

void ThreadBase::QueueUserAPC(PAPCFUNC pfnAPC, ULONG_PTR dwData) {
	if (::QueueUserAPC(pfnAPC, HandleAccess(_self), dwData))
		return;
	Throw(E_EXT_QueueUserAPC);				// No GetLastError value
}

void ThreadBase::QueueAPC() {
	try {
		QueueUserAPC(&APCProc, (DWORD_PTR)this);
	} catch (RCExc) {
	}
}
#	endif // !UCFG_WCE

#endif // UCFG_USE_PTHREADS

bool ThreadBase::Join(int ms) {
#if UCFG_USE_PTHREADS
	ASSERT(ms == INFINITE);
	void *r;	
	PthreadCheck(::pthread_join(m_tid, &r));
	m_exitCode = (DWORD)(DWORD_PTR)r;
	return true;
#else
	int r = ::WaitForSingleObject(HandleAccess(_self), ms);
	if (r == WAIT_TIMEOUT)
		return false;
	Win32Check(r != WAIT_FAILED);
	return true;
#endif
}

void ThreadBase::SignalStop() {
	m_bStop = true;
	if (Valid()) {
		if (m_bSleeping) {
#if UCFG_USE_PTHREADS
			PthreadCheck(::pthread_kill(m_tid, SIGUSR1));  //!!!? may be other signal
#elif !UCFG_WCE
			QueueAPC();
#endif
		}
	}
}

void ThreadBase::Stop() {
	SignalStop();
}

void ThreadBase::WaitStop() {
	if (Valid())
		Join();
}

/*
struct _AFX_THREAD_STARTUP {
	ThreadBase *m_pThread;
#ifdef WIN32
	AFX_MODULE_STATE *m_pModuleState;
#endif
};*/

void AFXAPI AfxTermThread(HINSTANCE hInstTerm) {
#if UCFG_EXTENDED
#	if !UCFG_WCE
	if (!hInstTerm)
		if (_AFX_THREAD_STATE* pThreadState = _afxThreadState.GetDataNA())
			pThreadState->m_pToolTip.reset();
#	endif
#endif
	if (_afxThreadData)
		_afxThreadData->DeleteValues(hInstTerm, false);		//!!! 
}

void AFXAPI AfxEndThread(UINT nExitCode, bool bDelete) {
#if UCFG_EXTENDED || UCFG_WIN32
	AFX_MODULE_THREAD_STATE* pState = AfxGetModuleThreadState();
	if (ThreadBase *pThread = pState->m_pCurrentThread) {
		pThread->OnEnd();
		if (bDelete)
			pThread->Delete();
		pState->m_pCurrentThread = 0;
	}
#else
	s_currentThread.Value = 0;
#endif
	AfxTermThread();
#if UCFG_USE_PTHREADS
	::pthread_exit((void*)(uintptr_t)nExitCode);
#elif UCFG_WCE
	::ExitThread(nExitCode);
#else
	_endthreadex(nExitCode);
#endif
}

void ThreadBase::OnEndThread(bool bDelete) {
	OnEnd();
	if (bDelete)
		Delete();
	AfxTermThread();
}

UInt32 ThreadBase::CppThreadThunk() {
	UInt32 exitCode = (UINT)E_FAIL; //!!!
	try {
		Run();
		exitCode = m_exitCode;
		OnEndThread(true);
	} catch (...) {
		ProcessExceptionInCatch();
	}
	TRC(0, "Thread Exited: " << hex << (int)ID << "\t" << Name);
	CCounterIncDec<ThreadBase, Interlocked>::Release(this);
	return exitCode;
}

#if UCFG_WCE

UINT ThreadBase::ThreaderFunction(LPVOID pParam) {
	__try {
//!!!R		_AFX_THREAD_STARTUP *pStartup = (_AFX_THREAD_STARTUP*)pParam;
		ThreadBase *pT = (ThreadBase*)pParam;
		return pT->CppThreadThunk();
	} __except (ProcessExceptionInFilter(GetExceptionInformation()), EXCEPTION_EXECUTE_HANDLER) {
		ProcessExceptionInExcept();
		return (UINT)E_FAIL;
	}	
}


#else

#if UCFG_USE_PTHREADS
	void * __cdecl ThreadBase::ThreaderFunction(void *pParam)
#else
	UINT ThreadBase::ThreaderFunction(LPVOID pParam)
#endif
{
	//!!!  set_terminate(Myterminate_handler);//!!!

	//!!!	__try
	//!!!	{
	//!!!  pThread->InitCOM();
	//!!!		bool bDelete = true;
	ThreadBase *pT = (ThreadBase*)pParam;;
#if UCFG_EXTENDED
	_AFX_THREAD_STATE* pThreadState = AfxGetThreadState();
	pThreadState->m_pModuleState = pT->m_pModuleStateForThread; //!!!pStartup->m_pModuleState;
	AFX_MODULE_STATE *pModuleState = AfxGetModuleState();
#ifdef _AFXDLL
	pT->m_pModuleState = pModuleState;
#endif
	pModuleState->m_thread->m_pCurrentThread = pT;
#else
	s_currentThread.Value = pT;
#endif
//!!!R	delete pStartup;

	DWORD r = pT->CppThreadThunk();

#if UCFG_USE_PTHREADS
	return (void*)(uintptr_t)r;
#else
	return r;
#endif

	/*!!!
	pT->TryExecute();
	DWORD exitCode = pT->ExitCode;
	AfxEndThread(exitCode, bDelete);
	return exitCode;*/
	/*!!!	}
	__except (MyExceptHandler(GetExceptionInformation()))
	{} */
	return 0;
}

#endif

void ThreadBase::Create(DWORD dwCreateFlags, size_t nStackSize
#ifdef WIN32
						, LPSECURITY_ATTRIBUTES lpSecurityAttrs
#endif
						)
{
#ifdef _DEBUG
	ASSERT(m_sig == 0xAB);		// should be in Heap
#endif

	ASSERT(!Valid());

	//!!!  m_pThreadState = AfxGetThreadState();

//!!!	_AFX_THREAD_STARTUP *pStartup = new _AFX_THREAD_STARTUP;


#ifdef WIN32
	m_pModuleStateForThread = AfxGetThreadState()->m_pModuleState;
#endif
	Interlocked::Increment(base::m_dwRef);
#if UCFG_USE_PTHREADS
	int rc = ::pthread_create(&m_tid, NULL, &ThreaderFunction, this);
	if (rc) {
		CCounterIncDec<ThreadBase, Interlocked>::Release(this);
		PthreadCheck(rc);
	}
#else
	HANDLE h;
	DWORD threadID;

#	if UCFG_WCE
		h = ::CreateThread(lpSecurityAttrs, nStackSize, (LPTHREAD_START_ROUTINE)&ThreaderFunction, this, dwCreateFlags, &threadID);
#	else
		h = (HANDLE)_beginthreadex(lpSecurityAttrs, (UINT)nStackSize, &ThreaderFunction, this, dwCreateFlags, (UINT*)&threadID); //!!!
#	endif
	if (!h || h==(HANDLE)-1) {
		CCounterIncDec<ThreadBase, Interlocked>::Release(this);
//!!!R		delete pStartup;
	}
	Attach(h, threadID);
#endif
}

void ThreadBase::Start(DWORD flags) {
	try {
		BeforeStart();
	} catch (RCExc) {
//!!!R		if (m_bAutoDelete)
//!!!R			delete this;
		throw;
	}
	if (m_owner) {
		EXT_LOCK (m_owner->m_cs) {
			m_owner->m_ar.push_back(this);
			m_owner->ExternalAddRef();
		}
	}
	Create(flags, StackSize);
}

CSeparateThreadThreader::CSeparateThreadThreader(CSeparateThread& st)
	:	ThreadBase(&st)
	,	m_st(st)
{
	m_bSleeping = true;
}

CSeparateThreadThreader::~CSeparateThreadThreader() {
	EXT_LOCK (m_st.m_cs) {
		m_st.m_pT = nullptr;
	}
}

void CSeparateThreadThreader::BeforeStart() {
	m_st.BeforeStart();
}

void CSeparateThreadThreader::Execute() {
	m_st.Execute();
}

void CSeparateThreadThreader::Stop() {
	ThreadBase::Stop();
	m_st.Stop();
}

CSeparateThread::~CSeparateThread() {
	StopChilds();
}

void CSeparateThread::Stop() {
//!!!R		if (m_pT)							// Endless recursion
//!!!R			m_pT->Stop();
}

void CSeparateThread::Create(DWORD flags) {
	(m_pT = new CSeparateThreadThreader(_self))->Start(flags);
}

#ifdef WIN32
CWinThread::CWinThread(CThreadRef *ownRef)
	:	ThreadBase(ownRef)
{
	m_pThreadParams = 0;
	m_pfnThreadProc = 0;
	CommonConstruct();
}

CWinThread::CWinThread(AFX_THREADPROC pfnThreadProc, LPVOID pParam) {
	m_pfnThreadProc = pfnThreadProc;
	m_pThreadParams = pParam;
	CommonConstruct();
}

void CWinThread::CommonConstruct() {
	m_hr = 0;
	m_lpfnOleTermOrFreeLib = 0;
}
#endif

#if UCFG_EXTENDED

void CWinThread::OnEnd() {
	if (m_lpfnOleTermOrFreeLib)
		m_lpfnOleTermOrFreeLib();
	ThreadBase::OnEnd();
}


MSG * AFXAPI AfxGetCurrentMessage() {
	return &AfxGetThreadState()->m_msgCur;
}

int CWinThread::ExitInstance() {
	return (int)AfxGetCurrentMessage()->wParam;
}

class CWndAttachKeeper {
	CWnd& m_wnd;
public:
	CWndAttachKeeper(CWnd& wnd, HWND hwnd)
		:	m_wnd(wnd)
	{
		m_wnd.Attach(hwnd);
	}

	~CWndAttachKeeper() {
		m_wnd.Detach();
	}
};

#endif

#if UCFG_EXTENDED


bool CWinThread::IsIdleMessage(MSG& msg) {
	if (msg.message == WM_MOUSEMOVE || msg.message == WM_NCMOUSEMOVE) {
		return true;
	}

	// WM_PAINT and WM_SYSTIMER (caret blink)
	return msg.message != WM_PAINT && msg.message != 0x0118;
}


bool CWinThread::DispatchThreadMessageEx(MSG& msg) {
	const AFX_MSGMAP* pMessageMap; pMessageMap = GetMessageMap();
	const AFX_MSGMAP_ENTRY* lpEntry;

	for (; pMessageMap->pfnGetBaseMap; pMessageMap=(*pMessageMap->pfnGetBaseMap)())
	{
		// Note: catch not so common but fatal mistake!!
		//      BEGIN_MESSAGE_MAP(CMyThread, CMyThread)
		ASSERT(pMessageMap != (*pMessageMap->pfnGetBaseMap)());

		if (msg.message < 0xC000)
		{
			// constant window message
			if ((lpEntry = AfxFindMessageEntry(pMessageMap->lpEntries,
				msg.message, 0, 0)) != NULL)
				goto LDispatch;
		}
		else
		{
			// registered windows message
			lpEntry = pMessageMap->lpEntries;
			while ((lpEntry = AfxFindMessageEntry(lpEntry, 0xC000, 0, 0)) != NULL)
			{
				UINT* pnID = (UINT*)(lpEntry->nSig);
				ASSERT(*pnID >= 0xC000);
				// must be successfully registered
				if (*pnID == msg.message)
					goto LDispatch;
				lpEntry++;      // keep looking past this one
			}
		}
	}
	return FALSE;

LDispatch:
	union MessageMapFunctions mmf;
	mmf.pfn_THREAD = 0;
	mmf.pfn = lpEntry->pfn;

	// always posted, so return value is meaningless

	(this->*mmf.pfn_THREAD)(msg.wParam, msg.lParam);
	return TRUE;
}

BOOL AfxInternalPreTranslateMessage(MSG* pMsg)
{
	//	ASSERT_VALID(this);

	CWinThread *pThread = AfxGetThread();
	if( pThread )
	{
		// if this is a thread-message, short-circuit this function
		if (pMsg->hwnd == NULL && pThread->DispatchThreadMessageEx(*pMsg))
			return TRUE;
	}

	// walk from target to main window
	CWnd* pMainWnd = AfxGetMainWnd();
	if (CWnd::WalkPreTranslateTree(pMainWnd->GetSafeHwnd(), pMsg))
		return TRUE;

	// in case of modeless dialogs, last chance route through main
	//   window's accelerator table
	if (pMainWnd != NULL)
	{
		CWnd* pWnd = CWnd::FromHandle(pMsg->hwnd);
		if (pWnd->GetTopLevelParent() != pMainWnd)
			return pMainWnd->PreTranslateMessage(pMsg);
	}

	return FALSE;   // no special processing
}

static BOOL AfxPreTranslateMessage(MSG* pMsg) {
	if (CWinThread *pThread = AfxGetThread())
		return pThread->PreTranslateMessage(*pMsg);
	else
		return AfxInternalPreTranslateMessage( pMsg );
}

static BOOL AfxInternalPumpMessage()
{
	_AFX_THREAD_STATE *pState = AfxGetThreadState();

	if (!::GetMessage(&(pState->m_msgCur), NULL, NULL, NULL))
	{
		/*!!!
		#ifdef _DEBUG
		TRACE(traceAppMsg, 1, "CWinThread::PumpMessage - Received WM_QUIT.\n");
		pState->m_nDisablePumpCount++; // application must die
		#endif
		*/
		// Note: prevents calling message loop things in 'ExitInstance'
		// will never be decremented
		return FALSE;
	}

#ifdef _DEBUG
	/*!!!
	if (pState->m_nDisablePumpCount != 0)
	{
	TRACE(traceAppMsg, 0, "Error: CWinThread::PumpMessage called when not permitted.\n");
	ASSERT(FALSE);
	}
	*/
#endif

#ifdef _DEBUG
	//!!!	_AfxTraceMsg(_T("PumpMessage"), &(pState->m_msgCur));
#endif

	// process this message

	if (pState->m_msgCur.message != WM_KICKIDLE && !AfxPreTranslateMessage(&(pState->m_msgCur)))
	{
		::TranslateMessage(&(pState->m_msgCur));
		::DispatchMessage(&(pState->m_msgCur));
	}
	return TRUE;
}

bool CWinThread::PreTranslateMessage(MSG& msg) {
	// if this is a thread-message, short-circuit this function
	if (msg.hwnd == NULL && DispatchThreadMessageEx(msg))
		return true;
	// walk from target to main window
	CWnd *pMainWnd = AfxGetMainWnd();
	if (CWnd::WalkPreTranslateTree(pMainWnd->GetSafeHwnd(), &msg))
		return true;
	// in case of modeless dialogs, last chance route through main
	//   window's accelerator table
	if (pMainWnd != NULL)
	{
		CWnd* pWnd = CWnd::FromHandle(msg.hwnd);
		if (pWnd->GetTopLevelParent() != pMainWnd)
			return pMainWnd->PreTranslateMessage(&msg);
	}
	return false;
}

bool CWinThread::PumpMessage() {
	return AfxInternalPumpMessage();
}

BOOL AFXAPI AfxPumpMessage() {
	if (CWinThread *pThread = AfxGetThread())
		return pThread->PumpMessage();
	else
		return AfxInternalPumpMessage();
}

int CWinThread::RunLoop() {
	bool bIdle = true;
	LONG lIdleCount = 0;
	_AFX_THREAD_STATE* pState = AfxGetThreadState();
	while (true)
	{
#if UCFG_GUI
		while (bIdle && !::PeekMessage(&(pState->m_msgCur), 0, 0, 0, PM_NOREMOVE))
			bIdle = OnIdle(lIdleCount++);
#endif
		do
		{
			//#ifndef NO_GUI //!!!
			if (!PumpMessage())
				return ExitInstance();
			//#endif

			if (IsIdleMessage(pState->m_msgCur))
			{
				bIdle = true;
				lIdleCount = 0;
			}
		}
		while (::PeekMessage(&(pState->m_msgCur), NULL, NULL, NULL, PM_NOREMOVE));
	}
}

void CWinThread::Execute() {
#if UCFG_GUI
	//!!!auto_ptr<CEvent> event2(pStartup->m_pEvent2);
	CWnd threadWnd;

	// forced initialization of the thread
	AfxInitThread();

	// thread inherits app's main window if not already set
	CWinApp* pApp = AfxGetApp();
	HWND hwnd = 0;
	if (pApp && !m_pMainWnd && pApp->m_pMainWnd->GetSafeHwnd())
	{
		// just attach the HWND
		hwnd = pApp->m_pMainWnd->m_hWnd;
		m_pMainWnd = &threadWnd;
	}
	CWndAttachKeeper wndAttachKeeper(threadWnd, hwnd);
	/*!!!  catch (RCExc e)
	{
	threadWnd.Detach();
	m_hr = e.HResult;
	//!!!      pStartup->m_hr = hr;
	//!!!      pStartup->m_event.Set();
	//!!!    bDelete = false; //!!!
	//    m_exitCode = (UINT)-1;
	throw;
	}
	*/
	//!!!    pStartup->m_event.Set();
	//!!!event2->Lock();
	// first -- check for simple worker thread
	if (m_pfnThreadProc)
		m_exitCode = (*m_pfnThreadProc)(m_pThreadParams);
	else if (!InitInstance())
		m_exitCode = ExitInstance();
	else
		RunLoop();
#endif
}

#endif


//!!!D Need for Win32s static long _afxTlsRef;

void AFXAPI AfxTlsAddRef() {
	//!!!D 	++_afxTlsRef;
}

void AFXAPI AfxTlsRelease() {
	//!!!D	if (_afxTlsRef == 0 || --_afxTlsRef == 0)
	{
		if (_afxThreadData) {
			_afxThreadData->~CThreadSlotData();
			_afxThreadData = NULL;
		}
	}
}

void CThreadSlotData::DeleteValues(CThreadData *pData, HINSTANCE hInst) {
	bool bDelete = true;
	for (int i = 1; i < pData->size(); i++) {
		if (!hInst || m_arSlotData[i].hInst == hInst) {
			// delete the data since hInst matches (or is NULL)
			delete (CNoTrackObject*)(*pData)[i];
			(*pData)[i] = 0;
		} else if ((*pData)[i])
			bDelete = false;
	}
	if (bDelete) {
		EXT_LOCK (m_criticalSection) {
			m_list.Remove(pData);
			m_tls.Value = 0;
		}
	}
}

#define SLOT_USED   0x01    // slot is allocated


void *CThreadSlotData::GetThreadValue(int nSlot) {
	ASSERT(nSlot && nSlot < m_nMax);
	ASSERT(m_arSlotData[nSlot].dwFlags & SLOT_USED);
	CThreadData *pData = (CThreadData*)(void*)m_tls.Value;
	if (!pData || nSlot >= pData->size())
		return 0;
	return (*pData)[nSlot];
}

CThreadLocalObject::~CThreadLocalObject() {
	if (m_nSlot != 0 && _afxThreadData)
		_afxThreadData->FreeSlot(m_nSlot);
	m_nSlot = 0;
}

void CThreadSlotData::FreeSlot(int nSlot) {
	EXT_LOCK (m_criticalSection) {
		ASSERT(nSlot && nSlot < m_nMax);
		ASSERT(m_arSlotData[nSlot].dwFlags & SLOT_USED);
		for (int i=0; i<m_list.size(); i++) {
			if (nSlot < m_list[i]->size()) {
				delete (CNoTrackObject*)(*m_list[i])[nSlot];
				(*m_list[i])[nSlot] = 0;
			}
		}
		m_arSlotData[nSlot].dwFlags &= ~SLOT_USED;
	}
}


int CThreadSlotData::AllocSlot() {
	EXT_LOCK (m_criticalSection) {
		int nAlloc = (int)m_arSlotData.size();
		int nSlot = m_nRover;
		if (nSlot >= nAlloc || (m_arSlotData[nSlot].dwFlags & SLOT_USED)) {
			// search for first free slot, starting at beginning
			for (nSlot = 1; nSlot < nAlloc && (m_arSlotData[nSlot].dwFlags & SLOT_USED); nSlot++)
				;

			// if none found, need to allocate more space
			if (nSlot >= nAlloc)
				m_arSlotData.resize(nSlot+1);
		}
		// adjust m_nMax to largest slot ever allocated
		if (nSlot >= m_nMax)
			m_nMax = nSlot+1;

		ASSERT(!(m_arSlotData[nSlot].dwFlags & SLOT_USED));
		m_arSlotData[nSlot].dwFlags |= SLOT_USED;
		// update m_nRover (likely place to find a free slot is next one)
		m_nRover = nSlot+1;

		return nSlot;   // slot can be used for FreeSlot, GetValue, SetValue
	}
}



void CThreadSlotData::SetValue(int nSlot, void *pValue) {
	ASSERT(nSlot && nSlot < m_nMax);
	ASSERT(m_arSlotData[nSlot].dwFlags & SLOT_USED);
	CThreadData *pData = (CThreadData*)(void*)m_tls.Value;
	if (!pData || nSlot >= pData->size() && pValue) {
		if (!pData) {
			pData = new CThreadData;
			EXT_LOCK (m_criticalSection) {
				m_list.Add(pData);
			}
		}
		pData->resize(m_nMax);
		m_tls.Value = pData;
	}
	(*pData)[nSlot] = pValue;
}


CNoTrackObject* CThreadLocalObject::GetDataNA() {
	if (m_nSlot == 0 || _afxThreadData == NULL)
		return NULL;

	CNoTrackObject* pValue =
		(CNoTrackObject*)_afxThreadData->GetThreadValue(m_nSlot);
	return pValue;
}


void *CNoTrackObject::operator new(size_t nSize) {
#ifdef WIN32
	void* p = ::LocalAlloc(LPTR, nSize);
#else
	void* p = malloc(nSize);
#endif
	if (!p)
		Throw(E_OUTOFMEMORY);
	return p;
}

void CNoTrackObject::operator delete(void* p) {
	if (p)
#ifdef WIN32
		::LocalFree(p);
#else
		free(p);
#endif
}

CTls::CTls() {
#if UCFG_USE_PTHREADS
	PthreadCheck(::pthread_key_create(&m_key, 0));
#else
	Win32Check((m_key = ::TlsAlloc()) != TLS_OUT_OF_INDEXES);
#endif
}

CTls::~CTls() {
#if UCFG_USE_PTHREADS
	PthreadCheck(::pthread_key_delete(m_key));
#else
	Win32Check(::TlsFree(m_key));
#endif
}

void CTls::put_Value(const void *p) {
#if UCFG_USE_PTHREADS
	PthreadCheck(::pthread_setspecific(m_key, p));
#else
	Win32Check(::TlsSetValue(m_key, (void*)p));
#endif
}



BYTE __afxThreadData[sizeof(CThreadSlotData)];

CThreadSlotData *_afxThreadData;


//!!!__declspec(nothrow)
CThreadSlotData::CThreadSlotData()
	:	m_nRover(1)
	,	m_nMax(0)
{
}

void CThreadSlotData::DeleteValues(HINSTANCE hInst, bool bAll) {
	EXT_LOCK (m_criticalSection) {
		if (!bAll) {
			CThreadData* pData = (CThreadData*)(void*)m_tls.Value;
			if (pData)
				DeleteValues(pData, hInst);
		} else
			for (int i=0; i<m_list.size(); i++) 
				DeleteValues(m_list[i].get(), hInst);
	}
}

CThreadSlotData::~CThreadSlotData() {
	DeleteValues(0, true);
}

void CThreadSlotData::AssignInstance(HINSTANCE hInst) {
	EXT_LOCK (m_criticalSection) {
		ASSERT(hInst != NULL);

		for (int i = 1; i < m_arSlotData.size(); i++) {
			if (m_arSlotData[i].hInst == NULL && (m_arSlotData[i].dwFlags & SLOT_USED))
				m_arSlotData[i].hInst = hInst;
		}
	}
}

void AFXAPI AfxInitLocalData(HINSTANCE hInst) {
	if (_afxThreadData)
		_afxThreadData->AssignInstance(hInst);
}

void AFXAPI AfxTermLocalData(HINSTANCE hInst, bool bAll) {
	if (_afxThreadData)
		_afxThreadData->DeleteValues(hInst, bAll);
}

void CThreadLocalObject::AllocSlot() {
	if (m_nSlot == 0) {
		if (_afxThreadData == NULL) { //!!!
			_afxThreadData = new(__afxThreadData) CThreadSlotData;
			ASSERT(_afxThreadData != NULL);
		}
		m_nSlot = _afxThreadData->AllocSlot();
		ASSERT(m_nSlot != 0);
	}
}

CNoTrackObject* CThreadLocalObject::GetData(CNoTrackObject* (AFXAPI * pfnCreateObject)()) {
	AllocSlot();			//!!! Normally Already Alloced
	CNoTrackObject* pValue = (CNoTrackObject*)_afxThreadData->GetThreadValue(m_nSlot);
	if (!pValue) {
		// allocate zero-init object
		pValue = (*pfnCreateObject)();

		// set tls data to newly created object
		_afxThreadData->SetValue(m_nSlot, pValue);
		ASSERT(_afxThreadData->GetThreadValue(m_nSlot) == pValue);
	}
	return pValue;
}

#ifdef WIN32
DWORD AFXAPI WaitWithMS(DWORD dwMS, HANDLE h0, HANDLE h1, HANDLE h2, HANDLE h3, HANDLE h4, HANDLE h5) {
	DWORD r = WAIT_FAILED;
	if (!h1)
		r = WaitForSingleObject(h0, dwMS);
	else {
		HANDLE ar[6] = {h0, h1, h2, h3, h4, h5};
		int i;
		for (i=0; i<6; i++)
			if (!ar[i])
				break;
		r = WaitForMultipleObjects(i, ar, FALSE, dwMS);
	}
	Win32Check(r != WAIT_FAILED);
	return r;
}

DWORD AFXAPI Wait(HANDLE h0, HANDLE h1, HANDLE h2, HANDLE h3, HANDLE h4, HANDLE h5) {
	return WaitWithMS(INFINITE, h0, h1, h2, h3, h4, h5);
}
#endif



void ReaderWriterLock::AcquireReaderLock() {
#if UCFG_USE_PTHREADS
	PthreadCheck(::pthread_rwlock_rdlock(&m_rwlock));
#else
	DWORD tid = ::GetCurrentThreadId();
	if (tid == m_tidWriter)
		return;
	DWORD_PTR tlsv = TlsValue+1;
	if (1 == tlsv) {
		EXT_LOCK (m_csWriters) {
			if (Interlocked::Increment(m_nActiveReaders) == 1)
				m_noReadsEvent.Reset();
		}
	}
	TlsValue = tlsv;
#endif
}

void ReaderWriterLock::ReleaseReaderLock() {
#if UCFG_USE_PTHREADS
	PthreadCheck(::pthread_rwlock_unlock(&m_rwlock));
#else
	DWORD tid = ::GetCurrentThreadId();
	if (tid == m_tidWriter)
		return;

	DWORD_PTR tlsv = TlsValue-1;
	TlsValue = tlsv;
	if (!tlsv && !Interlocked::Decrement(m_nActiveReaders))
		m_noReadsEvent.Set();
#endif
}

void ReaderWriterLock::AcquireWriterLock() {
#if UCFG_USE_PTHREADS
	PthreadCheck(::pthread_rwlock_wrlock(&m_rwlock));
#else
	DWORD tid = ::GetCurrentThreadId();
	if (tid == m_tidWriter) {
		m_nWriterDeep++;
		return;
	}
	m_csWriters.Lock();
	m_noReadsEvent.Lock();
	m_tidWriter = tid;
#endif
}

void ReaderWriterLock::ReleaseWriterLock() {
#if UCFG_USE_PTHREADS
	PthreadCheck(::pthread_rwlock_unlock(&m_rwlock));
#else
	if (m_nWriterDeep--)
		return;
	m_tidWriter = 0;
	m_csWriters.Unlock();
#endif
}

#ifdef _WIN32

#if UCFG_WIN32_FULL
void WorkItem::QueueAPC() {
	if (ExecutingThread)
		ExecutingThread->QueueAPC();
}
#endif

void WorkItem::Cancel() {
	if (!Finished) {
		Stop();
		if (ExecutingThread) {
			{
				EXT_LOCK (ExecutingThread->m_cs) {
					ExecutingThread->m_wiToCancel = this;
				}
			}
#if UCFG_WIN32_FULL
			ExecutingThread->QueueAPC();

#endif
		}
	}
}

PoolThread::PoolThread(ThreadPool& pool)
:	base(&pool.m_tr)
,	m_pool(pool)
{
}

void PoolThread::Execute() {
	Name = "PoolThread";

	while (true) {
		while (true) {
			if (m_bStop)
				return;
			ptr<WorkItem> wi;
			{
				EXT_LOCK (ThreadPool::I->m_cs) {
					if (!ThreadPool::I->m_queue.empty()) {
						wi = ThreadPool::I->m_queue.front();
						ThreadPool::I->m_queue.pop();
					}
				}
			}
			if (!wi)
				break;
			m_curWi = wi;
			wi->ExecutingThread = this;
			if (!wi->m_bStop) {
				try {
					wi->Execute();
				} catch (RCExc e) {
					wi->HResult = e.HResult;
				}
				wi->Finished = true;
				m_curWi = nullptr;
			}
		}
		EXT_LOCK (m_cs) {
			m_wiToCancel = nullptr;
		}
		int r = Wait(Handle(m_pool.m_evStop), Handle(ThreadPool::I->m_sem));
	}
}

void PoolThread::OnAPC() {
	if (m_wiToCancel == m_curWi)
		m_wiToCancel->OnAPC();
}


ThreadPool *ThreadPool::I;

ThreadPool::ThreadPool()
	:	m_evStop(FALSE, TRUE)
	,	m_sem(0, UCFG_POOL_THREADS)
{
	ASSERT(I == nullptr);
	I = this;

	for (int i=0; i<UCFG_POOL_THREADS; ++i)
		(new PoolThread(_self))->Start();
}

void ThreadPool::Ensure() {
	if (!I)
		AfxGetApp()->m_pThreadPool = new ThreadPool;
}



ptr<WorkItem> ThreadPool::QueueUserWorkItem(WorkItem *wi) {
	Ensure();
	EXT_LOCK (ThreadPool::I->m_cs) {
		ThreadPool::I->m_queue.push(wi);
		if (ThreadPool::I->m_queue.size() < UCFG_POOL_THREADS)
			ThreadPool::I->m_sem.Unlock();
	}
	return wi;
}
#endif  // _WIN32



} // Ext::

