/*###########################################################################################################################
# Copyright (c) 1997-2012 Ufasoft   http://ufasoft.com   mailto:support@ufasoft.com                                         #
#                                                                                                                           #
# This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License #
# as published by the Free Software Foundation; either version 3, or (at your option) any later version.                    #                                                          #
#                                                                                                                           #
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied        #
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.     #
#                                                                                                                           #
# You should have received a copy of the GNU General Public License along with this program;                                #
# If not, see <http://www.gnu.org/licenses/>                                                                                #
###########################################################################################################################*/

#pragma once

#include <shellapi.h>

namespace Ext {

class CDropFiles {
public:
	HDROP m_h;

	CDropFiles(HDROP h)
		:	m_h(h)
	{
	}

	~CDropFiles() {
		::DragFinish(m_h);
	}

	EXT_API CStringVector GetPaths();
};

class AFX_CLASS CServiceObject {
public:
	CServiceObject()
		:	m_handle(0)
	{}

	CServiceObject(const CServiceObject& s)
		:	m_handle(s.m_handle)
	{
		((CServiceObject&)s).m_handle = 0;
	}

	void operator=(CServiceObject& s) {
		Close();
		m_handle = s.m_handle;
		s.m_handle = 0;
	}

	virtual ~CServiceObject();
	void Close();

	SC_HANDLE m_handle;
};

class ServiceController : public CServiceObject {		//!!!
public:
	ServiceController() {
	}

	ServiceController(RCString serviceName, RCString machineName = nullptr);

	SERVICE_STATUS get_Status();
	DEFPROP_GET(SERVICE_STATUS,Status);

	void Delete();
	EXT_API void Start(const CStringVector& ar = CStringVector());
	SERVICE_STATUS ControlService(DWORD dwControl);
	SERVICE_STATUS Stop();
	SERVICE_STATUS Pause();
	SERVICE_STATUS Continue();
	SERVICE_STATUS Interrogate();
};

class CSCManager : public CServiceObject {	//!!!
	static const DWORD DEFAULT_ACCESS = MAXIMUM_ALLOWED;
	//!!!	const DWORD DEFAULT_ACCESS = SC_MANAGER_ALL_ACCESS;
public:
	CSCManager(RCString lpMachineName = nullptr, RCString lpDatabaseName = nullptr, DWORD dwDesiredAccess = DEFAULT_ACCESS);
	void Open(RCString lpMachineName = nullptr, RCString lpDatabaseName = nullptr, DWORD dwDesiredAccess = DEFAULT_ACCESS);
	EXT_API ServiceController CreateService(RCString serviceName, RCString displayName, DWORD dwDesiredAccess = DEFAULT_ACCESS,
		DWORD dwServiceType = SERVICE_WIN32_OWN_PROCESS, DWORD dwStartType = SERVICE_AUTO_START,
		DWORD dwErrorControl = SERVICE_ERROR_NORMAL, LPCTSTR lpBinaryPathName = 0, LPCTSTR lpLoadOrderGroup = 0,
		LPDWORD lpdwTagId = 0, LPCTSTR lpDependencies = 0, LPCTSTR lpServiceStartName = 0,
		LPCTSTR lpPassword = 0);
};

class AFX_CLASS CServiceHandler {
public:
	CServiceHandler(RCString serviceName, LPHANDLER_FUNCTION lpHandlerProc);
	void SetServiceStatus(SERVICE_STATUS& serviceStatus);
private:
	SERVICE_STATUS_HANDLE m_handle;
};

class AFX_CLASS CEnvironmentStrings {
public:
	LPTSTR m_p;

	CEnvironmentStrings()
		:	m_p(GetEnvironmentStrings())
	{
	}

	~CEnvironmentStrings() {
		Win32Check(FreeEnvironmentStrings(m_p));
	}
};

class CUnhandledExceptionFilter {
public:
	EXT_DATA static CUnhandledExceptionFilter *I;

	CUnhandledExceptionFilter();
	virtual ~CUnhandledExceptionFilter();
	void RestorePrev();

	virtual LONG Handle(EXCEPTION_POINTERS *ExceptionInfo) { return EXCEPTION_CONTINUE_SEARCH; }
protected:
	LPTOP_LEVEL_EXCEPTION_FILTER m_prev;
	CBool m_bSet;

private:
	static CCriticalSection s_cs;	

	static LONG CALLBACK UnhandledExceptionFilter(EXCEPTION_POINTERS* ExceptionInfo);
};

struct SFindData : public _finddata_t {
};

class EXTAPI CFileFind {
	SFindData m_fd;
	bool m_b;
public:
	intptr_t m_handle;

	CFileFind();
	CFileFind(RCString filespec);
	~CFileFind();
	void First(RCString filespec = nullptr);
	bool Next(SFindData& fd);
};

struct NamedPipeCreateInfo {
	String Name;
	CInt<DWORD> OpenMode, PipeMode;
	UInt32 MaxInstances;
	CPointer<SECURITY_ATTRIBUTES> Sa;

	NamedPipeCreateInfo()
		:	MaxInstances(PIPE_UNLIMITED_INSTANCES)
		,	OpenMode(PIPE_ACCESS_DUPLEX)
	{}
};

class NamedPipe : public File {
public:
	void Create(const NamedPipeCreateInfo& ci);

	void SetHandleState(DWORD mode) {
		Win32Check(::SetNamedPipeHandleState(Handle(*this), &mode, 0, 0));
	}

	bool Connect(LPOVERLAPPED ovl = NULL);
};

class StdHandle {
public:
	static HANDLE AFXAPI Get(DWORD nStdHandle);
	static HANDLE AFXAPI Set(DWORD nStdHandle, HANDLE h) { Win32Check(::SetStdHandle(nStdHandle, h)); }
};



} // Ext::
