/*###########################################################################################################################
# Copyright (c) 1997-2012 Ufasoft   http://ufasoft.com   mailto:support@ufasoft.com                                         #
#                                                                                                                           #
# This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License #
# as published by the Free Software Foundation; either version 3, or (at your option) any later version.                    #                                                          #
#                                                                                                                           #
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied        #
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.     #
#                                                                                                                           #
# You should have received a copy of the GNU General Public License along with this program;                                #
# If not, see <http://www.gnu.org/licenses/>                                                                                #
###########################################################################################################################*/

#include <el/ext.h>

#ifndef UCFG_BITCOIN_SOLO_MINING
#	define UCFG_BITCOIN_SOLO_MINING 0 //!!!
#endif

#include "miner.h"
using namespace Coin;


#if UCFG_BITCOIN_SOLO_MINING
#	include "../eng/eng.h"

class SoloWallet : public Object, public WalletBase {
	typedef WalletBase base;
public:
	ptr<CoinEng> m_pEng;
	Blob m_blobAddr;

	SoloWallet(ptr<CoinEng> eng)
		:	base(*eng)
		,	m_pEng(eng)
	{}

protected:

	pair<Blob, Blob> GetDefaultPublicKey() override {
		return make_pair(Blob(), m_blobAddr);
	}

	void OnProcessBlock(const Block& block) override {
		cerr << "\r " << Eng.ChainParams.Name << " Block " << block.Height << flush;
	}
	
	void OnProcessTx(const Tx& tx) override {}

	void OnEraseTx(const HashValue& hashTx) override {}

	void OnPeriodic() override {}

	void OnSetProgress(float v) override {}
	
	void OnChange() override {}

	bool IsFromMe(const Tx& tx) override { return false; }
};


class SoloMiner : public CoinDb {
public:
	vector<pair<String, String>> Chains;
	vector<ptr<SoloWallet>> SoloWallets;

	~SoloMiner() {
		for (int i=0; i<SoloWallets.size(); ++i)
			UnregisterForMining(SoloWallets[i]);
	}

	void StartChainEngs();
};

void SoloMiner::StartChainEngs() {
	DbWalletFilePath = nullptr;
	Load();

	for (int i=0; i<Chains.size(); ++i) {
		String saddr = Chains[i].first,
			chainname = Chains[i].second;
		ptr<SoloWallet> wallet = new SoloWallet(CoinEng::CreateObject(_self, chainname));
		Address addr(saddr);
		addr.CheckVer(wallet->Eng);
		wallet->m_blobAddr = addr;
		SoloWallets.push_back(wallet);
		wallet->Eng.Start();
		RegisterForMining(wallet);
	}
}

#endif	// UCFG_BITCOIN_SOLO_MINING


#include "bitcoin-sha256.h"

#if UCFG_WIN32 && UCFG_DEBUG && !defined(_AFXDLL)
#	pragma comment(lib, "ws2_32")
#	include <el/comp/crashdump.h>

class MyCrashDump : public CCrashDump {
	void ProcessDumpFile(RCString filename, EXCEPTION_POINTERS *ExceptionInfo) override {
		String dstFilename = System.ExeFilePath+".dmp";
		File::Delete(dstFilename);
		File::Move(filename, dstFilename);
		MessageBox::Show("Crash File " + dstFilename +" saved", "Unhandled exception!\n", (int)MessageBoxButtons::OK, MessageBoxIcon::Error);
	}
} g_crashDump;
#endif

class CMinerApp : public CConApp, public BitcoinMiner {
public:
	CThreadRef m_tr;
		

	CMinerApp() {
		CTrace::s_pOstream = &cerr;
		CTrace::s_nLevel = 3;

		FileDescription = VER_FILEDESCRIPTION_STR;
		SVersion			= VER_PRODUCTVERSION_STR;
		LegalCopyright = VER_LEGALCOPYRIGHT_STR;
		Url = VER_EXT_URL;

		MainBitcoinUrl = "http://127.0.0.1:8332";
	}

	void PrintUsage() {
		cerr << "Usage: bitcoin-miner {-options}"
#if UCFG_BITCOIN_SOLO_MINING
			" [bitcoin-address | {bitcoin-address chainname}]"
#endif
			    "\nOptions:\n"
				"  -a <seconds>        time between getwork requests 1..60, default 15\n"
				"  -A user-agent       Set custom User-agent string in HTTP header, default: Ufasoft bitcoin miner \n"
				"  -g yes|no           set \'no\' to disable GPU, default \'yes\'\n"
				"  -h                  this help\n"
				"  -i index|name       select device from Device List, can be used multiple times, default - all devices\n"
#if  UCFG_BITCOIN_LONG_POLLING
				"  -l yes|no           set \'no\' to disable Long-Polling, default \'yes\'\n"
#endif
				"  -o url              in form http://username:password@server.tld:port/path, by default http://127.0.0.1:8332\n"
				"  -t <threads>        Number of threads for CPU mining, by default is number of CPUs (Cores), 0 - disable CPU mining\n"			
#if UCFG_BITCOIN_THERMAL_CONTROL
				"  -T temperature      max temperature in Celsius degrees, default: " + Convert::ToString(BITCOIN_MAX_GPU_TEMPERATURE) + "\n"			
#endif
			    "  -v                  Verbose output\n"			
				"  -x type=host:port   Use HTTP or SOCKS proxy. Examples: -x http=127.0.0.1:3128, -x socks=127.0.0.1:1080\n"			
			 << endl;
	}

	void Execute() 	{
#if !UCFG_WIN32
		CPreciseTimeBase::s_pCurrent = nullptr;
#endif

#if UCFG_USE_LIBCURL && UCFG_WIN32
		CUsingSockets usingSockets;
#endif

		bool bMine = false;

		vector<String> selectedDevs;

		for (int arg; (arg = getopt(Argc, Argv, "a:A:g:hi:"
#if  UCFG_BITCOIN_LONG_POLLING
			"l:"
#endif
#if UCFG_BITCOIN_THERMAL_CONTROL
			"T:"
#endif
			"o:t:u:p:vx:")) != EOF;) {
			switch (arg)
			{
			case 'a':
				GetworkPeriod = atoi(optarg);
				if (GetworkPeriod<1 || GetworkPeriod>60) {
					cerr << "\nERROR: Time between getwork requests should be 1..60 seconds\n" << endl;
					PrintUsage();
					return;
				}
				break;
			case 'A':
				UserAgentString = optarg;
				break;
			case 'g':
				if (String(optarg) == "no")
					m_bTryGpu = false;
				break;
			case 'h':
				PrintUsage();
				return;
			case 'i':
				selectedDevs.push_back(optarg);
				bMine = true;
				break;
#if  UCFG_BITCOIN_LONG_POLLING
			case 'l':
				if (String(optarg) == "no")
					m_bLongPolling = false;
				break;
#endif
			case 'o':
				MainBitcoinUrl = optarg;
				bMine = true;
				break;
			case 'u':
				Login = optarg;
				break;
			case 'p':
				Password = optarg;
				break;
			case 't':
				ThreadCount = atoi(optarg);
				break;
#if UCFG_BITCOIN_THERMAL_CONTROL
			case 'T':
				MaxTemperature = (float)atoi(optarg);
				break;
#endif
			case 'v':
				++Verbosity;
				break;
			case 'x':
				ProxyString = optarg;
				break;
			default:
				PrintUsage();
				Throw(1);
			}
		}

		BitcoinMiner *miner = this;

#if UCFG_BITCOIN_SOLO_MINING
		SoloMiner soloMiner;
		if (optind < Argc) {
			if (optind == Argc-1)
				soloMiner.Chains.push_back(make_pair(String(Argv[optind]), String("bitcoin")));
			else {
				for (; optind < Argc-1; optind+=2)
					soloMiner.Chains.push_back(make_pair(String(Argv[optind]), String(Argv[optind+1])));
			}
			miner = &soloMiner;
			soloMiner.StartChainEngs();
		}
#endif

		CTrace::s_nLevel = (1 << Verbosity)-1;

		if (miner->ThreadCount<0 || miner->ThreadCount>32) {
			cerr << "Thread number should be 0..32" << endl;
			Throw(2);
		}

		miner->InitDevices();

		if (!bMine) {
			PrintUsage();
			
			cerr << "Device List:\n";
			for (int i=0; i<Devices.size(); ++i) {
				ComputationDevice& dev = *Devices[i];
				cerr << i+1 << ". " << dev.Name << "\t" << dev.Description << "\n";
			}
			cerr << endl;
			return;
		}

		if (!selectedDevs.empty()) {
			vector<ptr<ComputationDevice>> devices;
			for (int i=0; i<selectedDevs.size(); ++i) {
				String s = selectedDevs[i];
				if (int idx = atoi(s)) {
					if (idx > Devices.size())
						Throw(E_FAIL);
					devices.push_back(Devices[idx-1]);
				} else {
					for (int j=0; j<Devices.size(); ++j) {
						if (Devices[j]->Name == s) {
							devices.push_back(Devices[j]);
							break;
						}
					}
				}
			}
			Devices = devices;
		}

		cout << "Mining for " << MainBitcoinUrl << endl;		
		bool bComma = false;
		cout << "Using ";
		for (int i=0; i<Devices.size(); ++i) {
			if (SwapRet(bComma, true))
				cout << ", ";
			String name = Devices[i]->Name;
			cout << name;
			if (name == "CPU")
				cout << " (" << ThreadCount << " threads)";
		}
		cout << endl;

		miner->Start(&m_tr);

		while (!s_bSigBreak) {
			Thread::Sleep(2000);			

			cerr << "\r" << setprecision(3) << Speed/1000000 << " MHash/s     " << flush;
		}
		miner->Stop();		//!!! we can't stop LongPolling with easy libcurl

		double sec = (DateTime::UtcNow()-DtStart).TotalSeconds;
		cerr << "\nProcessed: " << EntireHashCount/1000000 << " Mhash, " << int(sec) << " s with average Rate: " << miner->Speed/1000000  << " MHash/s"
			<< "\nAccepted: " << AcceptedCount << ", average: " << AcceptedCount*60/sec << " shares/min"
			<< endl;
	}

	bool OnSignal(int sig) {
		CConApp::OnSignal(sig);
		return true;
	}
protected:

} theApp;



EXT_DEFINE_MAIN(theApp)



