/*###########################################################################################################################
# Copyright (c) 1997-2012 Ufasoft   http://ufasoft.com   mailto:support@ufasoft.com                                         #
#                                                                                                                           #
# This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License #
# as published by the Free Software Foundation; either version 3, or (at your option) any later version.                    #                                                          #
#                                                                                                                           #
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied        #
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.     #
#                                                                                                                           #
# You should have received a copy of the GNU General Public License along with this program;                                #
# If not, see <http://www.gnu.org/licenses/>                                                                                #
###########################################################################################################################*/

#include <el/ext.h>

#if UCFG_WIN32
#	include "resource.h"
#endif

#include <el/libext/ext-http.h>

#include "miner.h"

#include "bitcoin-sha256sse.h"

namespace Coin {
	
ComputationArchitecture *ComputationArchitecture::s_pChain;

class WebClientKeeper {
	WebClient*& m_cur;
	WebClient *m_prev;
public:
	WebClientKeeper(WebClient*& cur, WebClient& client)
		:	m_cur(cur)
		,	m_prev(cur)
	{
		cur = &client;
	}

	~WebClientKeeper() {
		m_cur = m_prev;
	}
};

void WorkerThreadBase::Stop() {
	base::Stop();
	Miner.m_evDataReady.Set();
}

void WorkerThreadBase::CompleteRound(UInt64 nHashes) {
	m_span += DateTime::UtcNow()-m_dtStartRound;
	m_nHashes += nHashes;

	Miner.OnRoundComplete();
}

class WorkerThread : public WorkerThreadBase {
	typedef WorkerThreadBase base;
public:
	WorkerThread(BitcoinMiner& miner, CThreadRef *tr)
		:	base(miner, tr)
	{}

	String GetDeviceName() override {
		return "CPU/"+CpuInfo().Name +"/"+ Convert::ToString(CpuInfo().Cpuid(1).EAX, 16);
	}
protected:
	void Execute() override ;

	friend class BitcoinMiner;
};

static regex s_rePair("\"(midstate|data|hash1|target|noncerange|firstnonce|lastnonce)\"\\s*:\\s*\"([^\"]+)");

BitcoinWorkData BitcoinWorkData::FromJson(RCString json) {
	BitcoinWorkData r;
	vector<String> ss = json.Split(",");
	for (int i=0; i<ss.size(); ++i) {
		String pair = ss[i];
		cmatch m;
		if (regex_search(pair.c_str(), m, s_rePair)) {
			String f = m[1],
				val = m[2];
			if (f == "midstate")
				r.Midstate = Blob::FromHexString(val);
			else if (f == "data")
				r.Data = Blob::FromHexString(val);
			else if (f == "hash1")
				r.Hash1 = Blob::FromHexString(val);
			else if (f == "target")
				r.Target = Blob::FromHexString(val);
			else if (f == "firstnonce")
				r.FirstNonce = Convert::ToUInt32(val, 16);
			else if (f == "lastnonce")
				r.LastNonce = Convert::ToUInt32(val, 16);
			else if (f == "noncerange") {
				Blob blob = Blob::FromHexString(val);
				if (blob.Size == 8) {
					r.FirstNonce = UInt32(ntohl(*(Int32*)blob.constData()));
					r.LastNonce = UInt32(ntohl(*(Int32*)(blob.constData()+4)));
				}
			}
		}
	}
	return r;
}

BitcoinWorkData BitcoinWorkData::GetTestData() {
#ifdef X_DEBUG//!!!D
	return FromJson( "{\"id\": \"26fffdfe-3890-11e0-8d10-fefd6d4ac335\", \"result\": "
		"{\"hash1\": \"00000000000000000000000000000000000000000000000000000000000000000000008000000000000000000000000000000000000000000000000000010000\", "
		"\"data\": \"000000010fa597b30f766c011dfd84e881360a50aa927954157ee47f00006c2600000000e0defbcaf28d04805b78b783d1336eea0300186e81543b2ace02c6a9"
		"A4B6D4E54EDA73561A0F61B1"
		"00000000000000800000000000000000000000000000000000000000000000000000000000000000000000000000000080020000\", "
		"\"midstate\": \"B7A18676FF37940FC22C297632ED98EE773455AE4E56D975F83C8EDFA328868D\", "
		"\"target\": \"ffffffffffffffffffffffffffffffffffffffffffffffffffffffff00000000\", "
		"\"firstnonce\":\"00000000\", \"lastnonce\":\"1FFFFFFF\"}, \"error\": null}");

#else
	return FromJson( "{\"id\": \"26fffdfe-3890-11e0-8d10-fefd6d4ac335\", \"result\": "
		"{\"hash1\": \"00000000000000000000000000000000000000000000000000000000000000000000008000000000000000000000000000000000000000000000000000010000\", "
		"\"data\": \"000000010fa597b30f766c011dfd84e881360a50aa927954157ee47f00006c2600000000e0defbcaf28d04805b78b783d1336eea0300186e81543b2ace02c6a9ac84f58b4d59b74d1b02855200000000000000800000000000000000000000000000000000000000000000000000000000000000000000000000000080020000\", "
		"\"midstate\": \"3c4249a8e1c54578a52d83c101e5c58ef52f030dee2e9d29b6949adfa69597ae\", "
		"\"target\": \"ffffffffffffffffffffffffffffffffffffffffffffffffffffffff00000000\", "
//valid too				"\"firstnonce\":\"30000000\", \"lastnonce\":\"3FFFFFFF\"}, \"error\": null}");
		"\"noncerange\":\"8D9CB6008D9DB600\"}, \"error\": null}");
#endif
}

void BitcoinWorkData::SetRollNTime(WebClient& wc) {
	String sRollNTime = wc.get_ResponseHeaders().Get("X-Roll-NTime");
	if (!!sRollNTime) {
		if (sRollNTime == "Y")
			RollNTime = 1;
		else if (atoi(sRollNTime))
			RollNTime = Convert::ToUInt16(sRollNTime);
	}
}

void CpuDevice::Start(BitcoinMiner& miner, CThreadRef *tr) {
	for (int i=0; i<miner.ThreadCount; ++i) {
		ptr<WorkerThread> t = new WorkerThread(miner, tr);
		t->Device = this;
		miner.Threads.push_back(t.get());
		t->Start();
	}
}

BitcoinMiner::BitcoinMiner()
	:	HashCount(0)
	,	AcceptedCount(0)
	,	NPAR(UCFG_BITCOIN_NPAR)
	,	Verbosity(0)
	,	GetworkPeriod(15)
	,	Login(nullptr)
	,	Password(nullptr)
	,	ThreadCount(Environment.ProcessorCount)
	,	MaxTemperature(BITCOIN_MAX_GPU_TEMPERATURE)
	,	GpuIdleMilliseconds(1)
	,	m_bTryGpu(true)
	,	m_bLongPolling(UCFG_BITCOIN_LONG_POLLING)
	,	m_dtSwichToMainUrl(DateTime::MaxValue)
	,	m_msWait(NORMAL_WAIT)
	,	m_minGetworkQueue(MIN_GETWORK_QUEUE)
	,	Speed(0)
	,	EntireHashCount(0)
{
	UserAgentString = String(MANUFACTURER);
#if UCFG_WIN32
	try {
		DBG_LOCAL_IGNORE_WIN32(ERROR_RESOURCE_TYPE_NOT_FOUND);	

		UserAgentString += " "+AfxGetCApp()->GetInternalName()+"/"+FileVersionInfo(System.ExeFilePath).ProductVersion;
	} catch (RCExc) {
	}
#else
	UserAgentString += " "+String(VER_INTERNALNAME_STR)+"/"+VER_PRODUCTVERSION_STR;
#endif
	UserAgentString += " (" + Environment.OSVersion.VersionString + ") ";
}

void BitcoinMiner::InitDevices() {
	if (!m_bInited) {
		m_bInited = true;

		ptr<ComputationDevice> dev = new CpuDevice;
		dev->Name = "CPU";
		dev->Description = String(CpuInfo().Name)+" "+Convert::ToString(Environment.ProcessorCoreCount)+" cores";
		Devices.push_back(dev);

		if (m_bTryGpu) {
			for (ComputationArchitecture *p=ComputationArchitecture::s_pChain; p; p=p->m_pNext) {
				p->FillDevices(_self);
			}
		}
	}
}

BitcoinWebClient BitcoinMiner::GetWebClient(WorkerThreadBase *wt) {
	BitcoinWebClient wc;
	wc.CacheLevel = RequestCacheLevel::BypassCache;
	wc.Credentials.UserName = Login;
	wc.Credentials.Password = Password;

	if (!ProxyString.IsEmpty())
		wc.Proxy = WebProxy::FromString(ProxyString);

	if (!UserAgentString.IsEmpty()) {
		wc.UserAgent = UserAgentString;
#if UCFG_BITCOIN_USERAGENT_INFO
		if (wt) {
			String devName = wt->GetDeviceName();
			if (!!devName) {
				static int s_cpus = Environment.ProcessorCount,
						   s_cores = Environment.ProcessorCoreCount;

				wc.UserAgent += " " + devName + " PC="+Convert::ToString(s_cpus)
										 + " CC="+Convert::ToString(s_cores);
				double sec = wt->m_span.TotalSeconds;
				if (sec > 1) {
					ostringstream os;
					os << " P=" << setprecision(3) << wt->m_nHashes/sec/1000000;
					wt->m_nHashes = 0;
					wt->m_span = TimeSpan();
					wc.UserAgent += os.str();
				}
				Int64 tsc = __rdtsc(),
					diff = tsc - SwapRet(wt->m_prevTsc, tsc);
				if (diff < 0 )
					wc.UserAgent += " TSCDiff="+Convert::ToString(diff);
			}
		}
#endif
	}
	wc.Headers.Set("X-Mining-Extensions", "hostlist longpoll noncerange rollntime switchto");
	return wc;
}

class LongPollingThread : public WorkerThreadBase {
	typedef WorkerThreadBase base;
public:
	String LongPollingUrl;

	LongPollingThread(BitcoinMiner& miner)
		:	base(miner, miner.m_tr)
	{}

	void Execute() override {
		Name = "LongPollingThread";

		try {
			Miner.CallNotifyRequest(_self, LongPollingUrl);
		} catch (RCExc e) {

			HRESULT hr = e.HResult;
			TRC(0, "Long Polling TIME-OUT");
			EXT_LOCK (Miner.m_csLongPollingThread) {
				Miner.m_LongPollingThread = nullptr;
			}
		}
	}
};

String BitcoinMiner::GetCurrentUrl() {
	String r;
	EXT_LOCK (m_csCurrentData) {

		r = BitcoinUrl;
	}
	return r;
}

static regex s_reBitcoinUrl("^(http://)?[^/]+(/)?");

void BitcoinMiner::CheckLongPolling(WebClient& wc) {
	EXT_LOCK (m_csLongPollingThread) {
		if (m_bLongPolling) {
			String longPollingPath = wc.get_ResponseHeaders().Get("X-Long-Polling");
			if (!!longPollingPath) {
				if (m_LongPollingThread)
					return;
				String url = GetCurrentUrl();
				cmatch m;
				if (regex_search(url.c_str(), m, s_reBitcoinUrl)) {
					LongPollingThread *t = new LongPollingThread(_self);
					int idx = -1;
					if (m[2].matched)
						idx = m.position(2);
					if (longPollingPath.Left(7) == "http://")
						t->LongPollingUrl = longPollingPath;
					else
						t->LongPollingUrl = (idx==-1 ? url : url.Left(idx)) + longPollingPath;
					m_LongPollingThread = t;
					t->Start();
				}
			} else if (m_LongPollingThread) {
				m_LongPollingThread->Stop();
				m_LongPollingThread = nullptr;
			}
		}
	}
}

static regex s_reUrlTtr("\"host\":\"([^\"]+)\",\"port\":(\\d+),\"ttr\":(\\d+)");

vector<SUrlTtr> GetUrlTtrs(RCString s) {
	vector<SUrlTtr> r;
	const char *p = s.c_str();
	for (cregex_iterator it(p, p+strlen(p), s_reUrlTtr), e; it!=e; ++it) {
		SUrlTtr urlTtr = { "http://"+String((*it)[1])+":"+String((*it)[2]) , Convert::ToInt32(String((*it)[3])) };
		r.push_back(urlTtr);
	}
	return r;
}


/*!!!R?
void BitcoinMiner::SetNewData(const BitcoinWorkData& wd) {
	if (wd.Midstate.Size != 32 || wd.Data.Size != 128 || wd.Target.Size != 32 || wd.Hash1.Size != 64)
		Throw(E_FAIL);

	DateTime now = DateTime::UtcNow();
	EXT_LOCK (m_csCurrentData) {
		TaskQueue.push_back(wd);
		m_dtNextGetWork = now + TimeSpan::FromSeconds(m_LongPollingThread ? 60 : GetworkPeriod);
		m_evDataReady.Set();
	}
}
*/

void BitcoinMiner::SetNewData(const BitcoinWorkData& wd, bool bClearOld) {
	if (wd.Midstate.Size != 32 || wd.Data.Size != 128 || wd.Target.Size != 32 || wd.Hash1.Size != 64)
		Throw(E_FAIL);

	DateTime now = DateTime::UtcNow();
	EXT_LOCK (m_csCurrentData) {
		if (bClearOld)
			TaskQueue.clear();
		else {
			DateTime dtObsolete = wd.Timestamp-TimeSpan::FromSeconds(60);
			while (TaskQueue.size() >= MAX_GETWORK_QUEUE || !TaskQueue.empty() && TaskQueue.front().Timestamp < dtObsolete)
				TaskQueue.pop_front();
		}

		TaskQueue.push_back(wd);
		for (int i=1; i<=wd.RollNTime; ++i) {
			BitcoinWorkData wd1 = wd;
			*reinterpret_cast<Int32*>(wd1.Data.data()+68) += htobe32(i);
			TaskQueue.push_back(wd1);
		}
		m_dtNextGetWork = now + TimeSpan::FromSeconds(m_LongPollingThread ? 60 : GetworkPeriod);
#ifdef X_DEBUG//!!!D
		m_dtNextGetWork = now+TimeSpan::FromSeconds(1000);
		this->m_minGetworkQueue = 1;
#endif
		m_evDataReady.Set();
		
	}
}

void BitcoinMiner::SetWebInfo(WebClient& wc) {
	DateTime now = DateTime::UtcNow();
	EXT_LOCK (m_csCurrentData) {
		if (now > m_dtSwichToMainUrl) {
			BitcoinUrl = MainBitcoinUrl;
			m_dtSwichToMainUrl = DateTime::MaxValue;
			TRACE_BITCOIN(cerr << "Switching back to " << BitcoinUrl  << endl);
		}

		String switchTo = wc.get_ResponseHeaders().Get("X-Switch-To");
		if (!!switchTo) {
			vector<SUrlTtr> urlTtrs = GetUrlTtrs(switchTo);
			if (!urlTtrs.empty()) {
				BitcoinUrl = urlTtrs[0].Url;
				m_dtSwichToMainUrl = now + TimeSpan::FromSeconds(urlTtrs[0].TtrMinutes*60);
				TRACE_BITCOIN(cerr << "Switching to " << BitcoinUrl  << endl);
			}
		}

		String hostList = wc.get_ResponseHeaders().Get("X-Host-List");
		if (!!hostList) {
			vector<SUrlTtr> urlTtrs = GetUrlTtrs(hostList);
			if (!urlTtrs.empty() && urlTtrs != HostList) {
				HostList = urlTtrs;
				HostIndex = 0;
				if (BitcoinUrl != HostList[0].Url) {
					TRACE_BITCOIN(cerr << "Switching to " << HostList[0].Url  << endl);
				}
				MainBitcoinUrl = BitcoinUrl = HostList[0].Url;
			}
		}
	}
}

void BitcoinMiner::CallNotifyRequest(WorkerThreadBase& wt, RCString url) {
	DBG_LOCAL_IGNORE_WIN32(ERROR_INTERNET_TIMEOUT);	
	DBG_LOCAL_IGNORE_WIN32(ERROR_INTERNET_OPERATION_CANCELLED);	

	BitcoinWebClient wc = GetWebClient(&wt);
	wc.Timeout = 180*1000;
	WebClientKeeper keeper(wt.CurrentWebClient, wc);

	String s = wc.DownloadString(url);
	TRC(0, "Long Polling returned");
	EXT_LOCK (m_csLongPollingThread) {
		m_LongPollingThread = nullptr;
	}
	if (BitcoinWorkData wd = BitcoinWorkData::FromJson(s)) {
		wd.SetRollNTime(wc);
		SetWebInfo(wc);
		CheckLongPolling(wc);  // should be before SetNewData() because SetNewData() uses m_LongPollingThread
		wd.Timestamp = DateTime::UtcNow();
		SetNewData(wd, true);
	}
}

BitcoinWorkData BitcoinMiner::GetWorkForThread(WorkerThreadBase& wt, UInt32 portion) {
	BitcoinWorkData wd;
	while (!m_bStop) {
		bool bGetWork = false;
		EXT_LOCK (m_csCurrentData) {
			for (CTaskQueue::iterator it=TaskQueue.begin(), e=TaskQueue.end(); it!=e; ++it) {
				Int64 req = Int64(portion) * UCFG_BITCOIN_NPAR,
					  avail = Int64(it->LastNonce)-Int64(it->FirstNonce)+1;
				if (portion != FULL_TASK) {
					Int64 delta = std::min(req, avail);
					wd = *it;
					wd.LastNonce = UInt32(wd.FirstNonce+delta-1);
					it->FirstNonce = wd.LastNonce+1;
					if (delta == avail)
						TaskQueue.erase(it);
					break;
				} else if (avail >= req) {
					wd = *it;
					TaskQueue.erase(it);
					break;
				}
			}
			bGetWork = TaskQueue.size() < m_minGetworkQueue;
		}
		if (bGetWork)
			m_evGetWork.Set();
		if (wd)
			break;
		m_evDataReady.Reset();
		m_evDataReady.Lock();
	}
	
/*!!!	UInt32 firstNonce = wd.FirstNonce+next*BitcoinMiner::NONCE_STEP,
			lastNonce = UInt32(firstNonce + Int64(BitcoinMiner::NONCE_STEP)*BitcoinSha256::NPAR/BitcoinSha256::NPAR - 1);    //  BitcoinSha256::NPAR can be non power of 2
	wd.FirstNonce = firstNonce;
	wd.LastNonce = lastNonce; */
	return wd;
}


BitcoinWorkData BitcoinMiner::GetWork(WebClient*& curWebClient) {
	String url = EXT_LOCKED(m_csCurrentData, BitcoinUrl);

	BitcoinWorkData wd;
	BitcoinWebClient wc = GetWebClient(nullptr);
	WebClientKeeper keeper(curWebClient, wc);
	try {
		wd.Timestamp = DateTime::UtcNow();
		if (wd = BitcoinWorkData::FromJson(wc.UploadString(url, "{\"method\": \"getwork\", \"params\": [], \"id\":0}"))) {
			wd.SetRollNTime(wc);
			SetWebInfo(wc);
			m_msWait = NORMAL_WAIT;
		}
#ifdef X_DEBUG//!!!D
		return BitcoinWorkData::FromJson( "{\"id\": \"26fffdfe-3890-11e0-8d10-fefd6d4ac335\", \"result\": "
		"{\"hash1\": \"00000000000000000000000000000000000000000000000000000000000000000000008000000000000000000000000000000000000000000000000000010000\", "
		"\"data\": \"000000010fa597b30f766c011dfd84e881360a50aa927954157ee47f00006c2600000000e0defbcaf28d04805b78b783d1336eea0300186e81543b2ace02c6a9"
		"89A648394EDA69FE1A0F61B1"
		"00000000000000800000000000000000000000000000000000000000000000000000000000000000000000000000000080020000\", "
		"\"midstate\": \"6F1F5CA2334BBEB9504D6527DE82FBC9F08A60EB93A82301800E7E939D1CD5DF\", "
		"\"target\": \"ffffffffffffffffffffffffffffffffffffffffffffffffffffffff00000000\", "
		"\"firstnonce\":\"00000000\", \"lastnonce\":\"FFFFFFFF\"}, \"error\": null}");
#endif
		CheckLongPolling(wc);  // should be before SetNewData() because SetNewData() uses m_LongPollingThread
		return wd;
	} catch (RCExc
#if UCFG_BITCOIN_TRACE
		e
#endif
		) {
			TRACE_BITCOIN(cerr << e.Message << endl);
			EXT_LOCK (m_csCurrentData) {
				if (HostIndex+1 < HostList.size()) {
					SUrlTtr& urlTtr = HostList[++HostIndex];
					BitcoinUrl = urlTtr.Url;
					m_dtSwichToMainUrl = urlTtr.TtrMinutes ? DateTime::UtcNow()+TimeSpan::FromSeconds(urlTtr.TtrMinutes*60) : DateTime::MaxValue;
					TRACE_BITCOIN(cerr << "Switching to " << BitcoinUrl  << endl);
					return wd;
				} else 
					HostIndex = 0;
			}
	}
	m_msWait = std::min(100000, m_msWait*2);
	return wd;
}

static regex s_reTrueResult("\"result\"\\s*:\\s*true");

bool BitcoinMiner::SubmitResult(WebClient*& curWebClient, const BitcoinWorkData& wd) {
	ostringstream sdata;
	sdata << wd.Data;
	String arg = "{\"params\":[\"" + String(sdata.str()) + "\"],\"method\":\"getwork\",\"id\":\"jsonrpc\"}";
	BitcoinWebClient wc = GetWebClient(nullptr);
	WebClientKeeper keeper(curWebClient, wc);
	String r = wc.UploadString(GetCurrentUrl(), arg);
	return regex_search(r.c_str(), s_reTrueResult);
}

bool BitcoinMiner::TestAndSubmit(WorkerThreadBase& wt, BitcoinWorkData& wd, UInt32 nonce) {
	*(UInt32*)(wd.Data.data()+64+12) = nonce;
	
	BitcoinSha256 sha256;
	sha256.PrepareData(wd.Midstate.constData(), wd.Data.constData()+64, wd.Hash1.constData());
	Blob hash = sha256.FullCalc();
	std::reverse(hash.data(), hash.data()+32);

	TRC(3, "Pre-result Nonce Found: " << hex << nonce << ", hash: " << hash);

	Blob rtarget = wd.Target;
	std::reverse(rtarget.data(), rtarget.data()+32);	

	if (memcmp(hash.constData(), rtarget.constData(), 32) < 0) {
		EXT_LOCK (m_csCurrentData) {
			WorksToSubmit.push_back(wd);
		}
		m_evGetWork.Set();
		return true;
	} else {
		TRC(2, "Wrong NONCE: " << hex << nonce);
		return false;
	}
}

String  BitcoinMiner::GetCalIlCode(bool bEvergreen) {
#ifdef X_DEBUG//!!!D
	FileStream fs("C:\\OUT\\Debug\\btc.il", FileMode::Open, FileAccess::Read);
	Blob blob(0, (size_t)fs.Length);
	fs.ReadBuffer(blob.data(), blob.Size);
	return Encoding::UTF8.GetChars(blob);
#endif
#if UCFG_WIN32
	CResID resId(bEvergreen ? ID_CAL_IL_EVERGREEN : ID_CAL_IL);
	Resource res(resId, RT_RCDATA);
	return Encoding::UTF8.GetChars(res);
#else
	Throw(E_NOTIMPL);
#endif
}

String BitcoinMiner::GetOpenclCode() {
#if UCFG_WIN32
	CResID resId(ID_OPENCL);
	Resource res(resId, RT_RCDATA);
	return Encoding::UTF8.GetChars(res);
#else
	Throw(E_NOTIMPL);
#endif
}

String BitcoinMiner::GetCudaCode() {
#if UCFG_WIN32
	CResID resId(ID_CUDA_PTX);
	Resource res(resId, RT_RCDATA);
	return Encoding::UTF8.GetChars(res);
#else
	Throw(E_NOTIMPL);
#endif
}

class GetWorkThread : public Thread {
	typedef Thread base;
public:
	BitcoinMiner& Miner;

	GetWorkThread(BitcoinMiner& miner, CThreadRef *tr)
		:	base(tr)
		,	Miner(miner)
		,	CurrentWebClient(0)
	{
		m_bSleeping = true;
	}

protected:
	WebClient *CurrentWebClient;

#if UCFG_WIN32
	void OnAPC() override {
		base::OnAPC();
		if (CurrentWebClient)
			CurrentWebClient->CurrentRequest->ReleaseFromAPC();
	}
#endif

	void Stop() override {
		base::Stop();
		Miner.m_evDataReady.Set();
		Miner.m_evGetWork.Set();
	}

	void Execute() override {
		Name = "GetWorkThread";

		DBG_LOCAL_IGNORE_WIN32(ERROR_INTERNET_TIMEOUT);
		DBG_LOCAL_IGNORE_WIN32(ERROR_HTTP_INVALID_SERVER_RESPONSE);
				
		deque<pair<DateTime, Int64>> deq;
		for (DateTime dt=Miner.DtStart; !m_bStop; ) {
			if (EXT_LOCKED(Miner.m_csCurrentData, dt > Miner.m_dtNextGetWork || Miner.TaskQueue.size() < Miner.m_minGetworkQueue)) {
				if (BitcoinWorkData wd = Miner.GetWork(CurrentWebClient)) {
					wd.Timestamp = dt = DateTime::UtcNow();
					Miner.SetNewData(wd);
					continue;
				}
			}
			
			Miner.m_evGetWork.Lock(Miner.m_msWait);

			while (true) {
				BitcoinWorkData wd;
				EXT_LOCK (Miner.m_csCurrentData) {
					if (Miner.WorksToSubmit.empty())
						break;
					wd = Miner.WorksToSubmit.front();
					Miner.WorksToSubmit.pop_front();
				}
				try {
					bool r = Miner.SubmitResult(CurrentWebClient, wd);
					if (r)
						Interlocked::Increment(Miner.AcceptedCount);

					BitcoinSha256 sha256;
					sha256.PrepareData(wd.Midstate.constData(), wd.Data.constData()+64, wd.Hash1.constData());
					Blob hash = sha256.FullCalc();
					std::reverse(hash.data(), hash.data()+32);

					String rejectReason = CurrentWebClient ? CurrentWebClient->get_ResponseHeaders().Get("X-Reject-Reason") : String(nullptr);
					ostringstream os;
					os << hash;
					String shash = os.str();		// GCC 4.4 don't allow EXT_STR
					if (!Miner.Verbosity)
						shash = shash.Substring(56, 8);
					TRACE_BITCOIN(cout << (isatty(fileno(stdout)) ? "\r" : "") << DateTime::Now() << " Result: " << shash << (r ? " accepted" : (" rejected "+(!rejectReason.IsEmpty() ? rejectReason : String()))) << endl);
				} catch (RCExc
		#if UCFG_BITCOIN_TRACE
					e
		#endif
					) {
					TRACE_BITCOIN(cerr << e.Message << endl);
				}
			}

			Int64 nProcessed = Int64(Interlocked::Exchange(Miner.HashCount, 0)) * UCFG_BITCOIN_NPAR;
			Miner.EntireHashCount += nProcessed;
			deq.push_back(make_pair(dt, nProcessed));
			double sec = ((dt = DateTime::UtcNow())-deq.front().first).TotalSeconds;
			if (sec > 1) {
				Int64 sum = 0;
				for (auto i=deq.begin(); i!=deq.end(); ++i)
					sum += i->second;
				Miner.Speed = float(sum/sec);
			}
			if (deq.size() > 20)
				deq.pop_front();
		}
	}
};


void BitcoinMiner::Start(CThreadRef *tr) {
	AFX_MANAGE_STATE(AfxGetStaticModuleState());

	if (!MainBitcoinUrl.IsEmpty()) {
		Uri url(MainBitcoinUrl);	
		if (Login.IsEmpty())
			Login = url.UserName;
		if (Password.IsEmpty())
			Password = url.Password;

		BitcoinUrl = MainBitcoinUrl;
	}
	m_tr = tr;
	InitDevices();
	if (!Threads.empty())
		return;
	DtStart = DateTime::UtcNow();
	m_evDataReady.Reset();
	m_evGetWork.Set();
	(GetWorkThread = new Coin::GetWorkThread(_self, tr))->Start();
	for (int i=0; i<Devices.size(); ++i)
		Devices[i]->Start(_self, tr);
}

void BitcoinMiner::Stop() {
	m_bStop = true;
	for (int i=0; i<Threads.size(); ++i)
		Threads[i]->Stop();
	ptr<Thread> longPollingThread = EXT_LOCKED(m_csLongPollingThread, m_LongPollingThread);
	if (longPollingThread)
		longPollingThread->Stop();
	if (GetWorkThread)
		GetWorkThread->Stop();
	for (int i=0; i<Threads.size(); ++i)
		Threads[i]->Join();
	if (GetWorkThread)
		GetWorkThread->Join();
	Threads.clear();
	GetWorkThread = nullptr;
	GpuThread = nullptr;
#if !UCFG_USE_LIBCURL
	if (longPollingThread)
		longPollingThread->Join();
#endif
//!!!R	m_LongPollingThread = nullptr;
	m_bStop = false;
}

void WorkerThread::Execute() {
	Name = "WorkerThread";

	Priority = THREAD_PRIORITY_LOWEST;

	while (!m_bStop) {

#ifdef X_DEBUG //!!!D
	BitcoinWorkData wd = BitcoinWorkData::GetTestData();
	/*!!!
	((UInt32*)wd.Data.data())[16+3] = wd.FirstNonce;
	BitcoinSha256 sha256;
	sha256.PrepareData(wd.Midstate.constData(), wd.Data.constData()+64, wd.Hash1.constData());
	Blob hash = sha256.FullCalc();*/

#else
		BitcoinWorkData wd = Miner.GetWorkForThread(_self, DevicePortion);
#endif

		if (!wd)
			return;
#ifdef X_DEBUG
		if ((wd.FirstNonce % BitcoinMiner::NONCE_STEP) != 0 ||
			((wd.LastNonce+1) % BitcoinMiner::NONCE_STEP) != 0
			|| wd.FirstNonce >= wd.LastNonce)
			cerr << "Invalid nonce numbers" << endl;
#endif


		BitcoinSha256 *bcSha;

		DECLSPEC_ALIGN(16) byte bufShaAlgo[sizeof(SseBitcoinSha256) + (16*(32*UCFG_BITCOIN_WAY+8)) + 256];		// max possible size with SSE buffers
#ifndef X_DEBUG //!!!D
		if (Miner.UseSse2())
			bcSha = new(bufShaAlgo) SseBitcoinSha256;
		else
#endif
			bcSha = new(bufShaAlgo) BitcoinSha256;

		bcSha->PrepareData(wd.Midstate.constData(), wd.Data.constData()+64, wd.Hash1.constData());

		StartRound();
		for (UInt32 nonce=wd.FirstNonce, end=wd.LastNonce+1; nonce!=end;) {
			if (bcSha->FindNonce(nonce)) {
				BitcoinSha256 sha256;
				sha256.PrepareData(wd.Midstate.constData(), wd.Data.constData()+64, wd.Hash1.constData());
				while (true) {
					if (!Miner.TestAndSubmit(_self, wd, nonce)) {
						TRACE_BITCOIN(cerr << "Found NONCE not accepted by Target" << endl);
					}
/*!!!R					bool b = sha256.FindNonce(nonce); //!!!D
					cout << "sha256.FindNonce(nonce)= " << b << "  " << hex << nonce << endl;  //!!!D
					*/
					if (!(++nonce % UCFG_BITCOIN_NPAR) || !sha256.FindNonce(nonce))
						break;
				}
			}				

			Interlocked::Increment(Miner.HashCount);
			if (m_bStop)
				break;
		}
		CompleteRound(wd.LastNonce-wd.FirstNonce+1);
	}
}

#ifdef _AFXDLL 
	 CWinApp theMiner;
#endif


} // Coin::

